/*
 * Copyright (c) 1997 Massachusetts Institute of Technology
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to use, copy, modify, and distribute the Software without
 * restriction, provided the Software, including any modified copies made
 * under this license, is not distributed for a fee, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE MASSACHUSETTS INSTITUTE OF TECHNOLOGY BE LIABLE
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
 * CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 * 
 * Except as contained in this notice, the name of the Massachusetts
 * Institute of Technology shall not be used in advertising or otherwise
 * to promote the sale, use or other dealings in this Software without
 * prior written authorization from the Massachusetts Institute of
 * Technology.
 *  
 */

#include <stdio.h>
#include <stdlib.h>

#include <math.h>

#include "bench_utils.h"
#include "bench_1d_protos.h"
#include "bench_3d_protos.h"
#include "bench_ffts.h"

extern int check_prime_factors(int n, int maxprime);

/* Skeleton for typical routine to perform benchmark for an fft:

void do_foo_fft(int rank, int *n, int *n_rev, int N, short is_power_of_two,
                FFTW_COMPLEX *arr, FFTW_COMPLEX *work,
                int size_arr, int size_work,
                short compute_accuracy, factor_type allowed_factors)
{
     if (rank not okay) return;
     if (N != 0) {
          check if N is okay for foo...
          init foo fft...
     }
     if (N is okay for foo...)
          DO_BENCHMARK_ND("Foo",rank,n,N,...)
     else
          SKIP_BENCHMARK("Foo could not handle this N because...")
}

This routine should be called from bench_1d/bench_3d, inserted in
alphabetical order by "Foo".

For Fortran FFTs, the body of the do_foo_fft function should be bracketed
by #ifdef USE_FORTRAN ... #endif directives.

*/

void do_arndt_ffts(int rank, int *n, int *n_rev, int N, short is_power_of_two,
		   FFTW_COMPLEX *arr, FFTW_COMPLEX *work,
		   int size_arr, int size_work,
		   short compute_accuracy, factor_type allowed_factors)
{
     int log2N = 0;
     
     if (rank != 1) return;
     if (N > 0 && is_power_of_two) {
	  int n1 = N;
	  while (n1 % 2 == 0) {
	       n1 /= 2;
	       log2N += 1;
	  }
     }
     if (is_power_of_two)
	  DO_BENCHMARK_ND("Arndt DIF",
			  rank, n, N, arr, arr,
			  dif4_fft((FFTW_REAL*)arr, 
				   (FFTW_REAL*)arr + N, log2N, -1),
			  1.0, -1, 0,
			  dif4_fft((FFTW_REAL*)arr, 
				   (FFTW_REAL*)arr + N, log2N, +1),
			  1.0/N,
			  compute_accuracy)
     else if (allowed_factors != NON_POWERS_OF_TWO_ONLY)
	  SKIP_BENCHMARK("not a power of 2");
     if (is_power_of_two)
	  DO_BENCHMARK_ND("Arndt DIT",
			  rank, n, N, arr, arr,
			  dit4_fft((FFTW_REAL*)arr, 
				   (FFTW_REAL*)arr + N, log2N, -1),
			  1.0, -1, 0,
			  dit4_fft((FFTW_REAL*)arr, 
				   (FFTW_REAL*)arr + N, log2N, +1),
			  1.0/N,
			  compute_accuracy)
     else if (allowed_factors != NON_POWERS_OF_TWO_ONLY)
	  SKIP_BENCHMARK("not a power of 2");
     if (is_power_of_two)
	  DO_BENCHMARK_ND("Arndt Split-Radix",
			  rank, n, N, arr, arr,
			  duhamel_fft((FFTW_REAL*)arr, 
				      (FFTW_REAL*)arr + N, log2N, -1),
			  1.0, -1, 0,
			  duhamel_fft((FFTW_REAL*)arr, 
				      (FFTW_REAL*)arr + N, log2N, +1),
			  1.0/N,
			  compute_accuracy)
     else if (allowed_factors != NON_POWERS_OF_TWO_ONLY)
	  SKIP_BENCHMARK("not a power of 2");
     if (is_power_of_two)
	  DO_BENCHMARK_ND("Arndt 4-step",
			  rank, n, N, arr, arr,
			  four_step_fft((FFTW_REAL*)arr, 
					(FFTW_REAL*)arr + N, log2N, -1),
			  1.0, -1, 0,
			  four_step_fft((FFTW_REAL*)arr, 
					(FFTW_REAL*)arr + N, log2N, +1),
			  1.0/N,
			  compute_accuracy)
     else if (allowed_factors != NON_POWERS_OF_TWO_ONLY)
	  SKIP_BENCHMARK("not a power of 2");
}

void do_bailey_fft(int rank, int *n, int *n_rev, int N, short is_power_of_two,
		   FFTW_COMPLEX *arr, FFTW_COMPLEX *work,
		   int size_arr, int size_work,
		   short compute_accuracy, factor_type allowed_factors)
{
#ifdef USE_FORTRAN
     if (rank != 1) return;
     if (is_power_of_two && sizeof(FFTW_REAL) == sizeof(double)) {
	  int n1 = N, m = 0, is1 = -1, is2 = +1;
	  FFTW_REAL  *u = (FFTW_REAL *) (work + N);
	  
	  if (N != 0) {
	       while (n1 % 2 == 0) {
		    ++m;
		    n1 /= 2;
	       }
	       FORTRANIZE(mpinix,MPINIX)(&m,u);
	  }
	  DO_BENCHMARK_ND("Bailey",
			  rank, n, N, arr, arr,
			  FORTRANIZE(mpcfft,MPCFFT)(&is1,&m,arr,work,u),
			  1.0, -1, 0,
			  FORTRANIZE(mpcfft,MPCFFT)(&is2,&m,arr,work,u),
			  1.0/N,
			  compute_accuracy);	  
     }
     else if (sizeof(FFTW_REAL) == sizeof(double) &&
	      allowed_factors != NON_POWERS_OF_TWO_ONLY)
	  SKIP_BENCHMARK("not a power of 2");
#endif
}

void do_beauregard_fft(int rank, int *n, int *n_rev, int N, 
		       short is_power_of_two,
		       FFTW_COMPLEX *arr, FFTW_COMPLEX *work,
		       int size_arr, int size_work,
		       short compute_accuracy, factor_type allowed_factors)
{
     if (rank != 1) return;
     if (is_power_of_two) {
	  int log2n = 0;
	  
	  if (N != 0) {
	       int np = N;
	       while (np > 1) {
		    np >>= 1;
		    log2n += 1;
	       }
	  }
	  
	  DO_BENCHMARK_ND("Beauregard",
			  rank, n, N, arr, arr,
			  BeauregardFFT(arr, log2n, 0),
			  1.0, -1, 1,
			  BeauregardFFT(arr, log2n, 1),
			  1.0,
			  compute_accuracy);
     }
     else if (allowed_factors != NON_POWERS_OF_TWO_ONLY)
	  SKIP_BENCHMARK("not a power of 2");
}

void do_bergland_fft(int rank, int *n, int *n_rev, int N, 
		     short is_power_of_two,
		     FFTW_COMPLEX *arr, FFTW_COMPLEX *work,
		     int size_arr, int size_work,
		     short compute_accuracy, factor_type allowed_factors)
{
     if (rank != 1) return;
     if (is_power_of_two && N <= (1L << 20))
	  DO_BENCHMARK_ND("Bergland",
			  rank, n, N, arr, arr,
			  bergland_fft1d(arr,N),
			  1.0, +1, 1,
			  0,
			  1.0/N,
			  -compute_accuracy)
     else if (is_power_of_two && N > (1L << 20))
	  SKIP_BENCHMARK("Bergland doesn't work for N > 2^20")
     else if (allowed_factors != NON_POWERS_OF_TWO_ONLY)
	  SKIP_BENCHMARK("not a power of 2");
}

void do_brenner_fft(int rank, int *n, int *n_rev, int N, short is_power_of_two,
		    FFTW_COMPLEX *arr, FFTW_COMPLEX *work,
		    int size_arr, int size_work,
		    short compute_accuracy, factor_type allowed_factors)
{
#ifdef USE_FORTRAN
     if (rank != 1) return;
     if (sizeof(FFTW_REAL) == sizeof(double)) {
	  int n1 = N, n2, rank = 1, iform = 1, is1 = -1, is2 = +1;
	  
	  /* There seems to be a bug in Brenner for N=3*2^m.  
	     Any other N seems to work fine!  Ugh... */
	  n2 = N;
	  if (N != 0) while (n2 % 2 == 0) n2 /= 2;
	  if (n2 != 3)
	       DO_BENCHMARK_ND("Brenner",
			       rank, n, N, arr, arr,
			       FORTRANIZE(brennerfft,
					  BRENNERFFT)(arr,&n1,&rank,&is1,
						       &iform,work),
			       1.0, -1, 1,
			       FORTRANIZE(brennerfft,
					  BRENNERFFT)(arr,&n1,&rank,&is2,
						       &iform,work),
			       1.0/N,
			       compute_accuracy)
	  else
	       SKIP_BENCHMARK("Brenner has a bug for N=3*2^m");
     }
#endif
}

void do_burrus_fft(int rank, int *n, int *n_rev, int N, short is_power_of_two,
		   FFTW_COMPLEX *arr, FFTW_COMPLEX *work,
		   int size_arr, int size_work,
		   short compute_accuracy, factor_type allowed_factors)
{
#ifdef USE_FORTRAN
if (sizeof(FFTW_REAL) == sizeof(double)) {
     if (rank != 1) return;
     if (is_power_of_two) {
	  int n1 = N, m = 0, is1 = -1, is2 = +1;
	  FFTW_REAL  *u = (FFTW_REAL *) (work + N);
	  
	  if (N != 0) {
	       while (n1 % 2 == 0) {
		    ++m;
		    n1 /= 2;
	       }
	  }
	  DO_BENCHMARK_ND("Burrus",
			  rank, n, N, arr, arr,
			  FORTRANIZE(burruscfft,BURRUSCFFT)
			       ((FFTW_REAL*)arr,(FFTW_REAL*)arr + N,
				&N,&m,&is2),
			  1.0, -1, 0,
			  FORTRANIZE(burruscfft,BURRUSCFFT)
			       ((FFTW_REAL*)arr,(FFTW_REAL*)arr + N,
				&N,&m,&is1),
			  1.0,
			  compute_accuracy);	  
     }
     else if (allowed_factors != NON_POWERS_OF_TWO_ONLY)
	  SKIP_BENCHMARK("not a power of 2");
}
#endif
}

void do_cwp_fft(int rank, int *n, int *n_rev, int N, short is_power_of_two,
                FFTW_COMPLEX *arr, FFTW_COMPLEX *work,
                int size_arr, int size_work,
                short compute_accuracy, factor_type allowed_factors)
{
     int np = N;
     
     if (rank != 1) return;

     if (N != 0) np = npfa(N);
     
     if (np >= N && np <= size_arr)
	  DO_BENCHMARK_ND("CWP (min N)",
			  1, &np, N, arr, arr,
			  pfacc(-1, np, arr), 1.0, -1, -1,
			  pfacc(+1, np, arr), 1.0/np,
			  compute_accuracy)
     else if (np < N)
	  SKIP_BENCHMARK("this transform size is too big for CWP")
     else
	  SKIP_BENCHMARK("unknown error...array isn't big enough for CWP!")
     
     if (N != 0) np = npfao(N, size_arr);
     
     if (np >= N && np <= size_arr)
	  DO_BENCHMARK_ND("CWP (best N)",
			  1, &np, N, arr, arr,
			  pfacc(-1, np, arr), 1.0, -1, -1,
			  pfacc(+1, np, arr), 1.0/np,
			  compute_accuracy)
     else if (np < N)
	  SKIP_BENCHMARK("this transform size is too big for CWP")
     else
	  SKIP_BENCHMARK("unknown error...array isn't big enough for CWP!")
}


void do_edelblute_fft(int rank, int *n, int *n_rev, int N,
		      short is_power_of_two,
		      FFTW_COMPLEX *arr, FFTW_COMPLEX *work,
		      int size_arr, int size_work,
		      short compute_accuracy, factor_type allowed_factors)
{
     if (rank != 1) return;
     if (is_power_of_two && (N == 0 || N > 2))
	  DO_BENCHMARK_ND("Edelblute",
			  rank, n, N, arr, arr,
			  fft_duhamel((FFTW_REAL*)arr, 
				      (FFTW_REAL*)arr + N, N),
			  1.0, -1, 0,
			  0,
			  1.0/N,
			  -compute_accuracy)
     else if (N == 2 || N == 1)
	  SKIP_BENCHMARK("Edelblute can't handle N <= 2")
     else if (allowed_factors != NON_POWERS_OF_TWO_ONLY)
	  SKIP_BENCHMARK("not a power of 2");
}

void do_fftpack_fft(int rank, int *n, int *n_rev, int N, short is_power_of_two,
		    FFTW_COMPLEX *arr, FFTW_COMPLEX *work,
		    int size_arr, int size_work,
		    short compute_accuracy, factor_type allowed_factors)
{
#ifdef USE_FORTRAN
     int n1 = N;
     if (rank != 1) return;
     if (sizeof(FFTW_REAL) == sizeof(double)) {
	  if (N != 0) FORTRANIZE(dcffti,DCFFTI)(&n1, work);
	  DO_BENCHMARK_ND("FFTPACK",
			  rank, n, N, arr, arr,
			  FORTRANIZE(dcfftf,DCFFTF)(&n1, arr, work),
			  1.0, -1, 1,
			  FORTRANIZE(dcfftb,DCFFTB)(&n1, arr, work),
			  1.0/N,
			  compute_accuracy);
     }
     else if (sizeof(FFTW_REAL) == sizeof(float)) {
	  if (N != 0) FORTRANIZE(scffti,SCFFTI)(&n1, work);
	  DO_BENCHMARK_ND("FFTPACK",
			  rank, n, N, arr, arr,
			  FORTRANIZE(scfftf,SCFFTF)(&n1, arr, work),
			  1.0, -1, 1,
			  FORTRANIZE(scfftb,SCFFTB)(&n1, arr, work),
			  1.0/N,
			  compute_accuracy);
     }
#endif
}


void do_fftpack_f2c_fft(int rank, int *n, int *n_rev, int N, 
			short is_power_of_two,
			FFTW_COMPLEX *arr, FFTW_COMPLEX *work,
			int size_arr, int size_work,
			short compute_accuracy, factor_type allowed_factors)
{
     if (rank != 1) return;
     if (N != 0) fftpack_cffti(N, work);
     DO_BENCHMARK_ND("FFTPACK (f2c)",
		     rank, n, N, arr, arr,
		     fftpack_cfftf(N, arr, work),
		     1.0, -1, 1,
		     fftpack_cfftb(N, arr, work),
		     1.0/N,
		     compute_accuracy);     	  

}

short do_fftw_1d_fft(int rank, int *n, int *n_rev, int N, 
		     short is_power_of_two,
		     FFTW_COMPLEX *arr, FFTW_COMPLEX *work,
		     int size_arr, int size_work,
		     short compute_accuracy, factor_type allowed_factors)
{
     fftw_plan p = 0, p_inv = 0, p_est = 0;
     int wis_flags = 0;

     #ifdef FFTW_HAS_WISDOM /* if we are using a version of FFTW with the
			       "wisdom" feature, then take advantage of it */
/*     wis_flags = FFTW_USE_WISDOM;  */
     #endif
     
     if (rank != 1) return 0;

     /* don't bother creating plan if which_fft != "FFTW" */
     if (which_fft[0] == 0 || !strcmp(which_fft,"FFTW")) {
	  if (N != 0) {
	       p = fftw_create_plan(N,FFTW_FORWARD,FFTW_MEASURE|wis_flags);
	       p_inv = fftw_create_plan(N,FFTW_BACKWARD,FFTW_ESTIMATE);
	       if (!p || !p_inv) {
		    fprintf(bench_log_file,
			    "\n\nError creating FFTW plan!\n");
		    printf("\n\nError creating FFTW plan!\n");
		    return 1;
	       }
#ifdef FFTW_HAS_FPRINT_PLAN
	       if (bench_log_file) {
		    fprintf(bench_log_file,"\nFFTW_MEASURE ");
		    fftw_fprint_plan(bench_log_file,p);
		    fprintf(bench_log_file,"\n");
	       }
#endif
	  }
	  
	  DO_BENCHMARK_ND("FFTW",
			  1, n, N, arr, work,
			  fftw(p,1,arr,1,0,work,1,0), 
			  1.0, FFTW_FORWARD, 1,
			  fftw(p_inv,1,work,1,0,arr,1,0), 1.0/N,
			  compute_accuracy);

	  if (N != 0)
	       fftw_destroy_plan(p);
     }

     /* don't bother creating plan if which_fft != "FFTW_ESTIMATE" */
     if (which_fft[0] == 0 || !strcmp(which_fft,"FFTW_ESTIMATE")) {
	  if (N != 0) {
	       if (!p_inv)
		    p_inv = fftw_create_plan(N,FFTW_BACKWARD,FFTW_ESTIMATE);
	       p_est = fftw_create_plan(N,FFTW_FORWARD,FFTW_ESTIMATE);
	       if (!p_est || !p_inv)
		    return 666;
#ifdef FFTW_HAS_FPRINT_PLAN
	       if (bench_log_file) {
		    fprintf(bench_log_file,"\nFFTW_ESTIMATE ");
		    fftw_fprint_plan(bench_log_file,p_est);
		    fprintf(bench_log_file,"\n");
	       }
#endif
	  }
	  
	  DO_BENCHMARK_ND("FFTW_ESTIMATE",
			  1, n, N, arr, work,
			  fftw(p_est,1,arr,1,0,work,1,0), 
			  1.0, FFTW_FORWARD, 1,
			  fftw(p_inv,1,work,1,0,arr,1,0), 1.0/N,
			  compute_accuracy);
     }

     if (p_inv)
	  fftw_destroy_plan(p_inv);
     if (p_est)
	  fftw_destroy_plan(p_est);

     return 0;
}

void do_frigo_fft(int rank, int *n, int *n_rev, int N, 
		  short is_power_of_two,
		  FFTW_COMPLEX *arr, FFTW_COMPLEX *work,
		  int size_arr, int size_work,
		  short compute_accuracy, factor_type allowed_factors)
{
     int factors[40];
 
     if (rank != 1) return;
     if (N != 0) cilk_many_fft_init(N, work, factors);
     DO_BENCHMARK_ND("Frigo-old",
		     1, n, N, arr, work,
		     cilk_fft_aux(N, arr, work, factors, 
				  work + N, N),
		     1.0, -1, 1,
		     0, 1.0,
		     -compute_accuracy);
}

void do_temperton_fft(int rank, int *n, int *n_rev, int N, 
		 short is_power_of_two,
		 FFTW_COMPLEX *arr, FFTW_COMPLEX *work,
		 int size_arr, int size_work,
		 short compute_accuracy, factor_type allowed_factors)
{
     if (rank != 1) return;
     if (check_prime_factors(N,5) && (N == 0 || 0 == setgpfa(work,N)))
	  DO_BENCHMARK_ND("Temperton (f2c)",
			  1, n, N, arr, arr,
			  gpfa((FFTW_REAL*)arr, (FFTW_REAL*)arr + 1, 
			       work, 2, 2 * N, N, 1, -1),
			  1.0, -1, 1,
			  gpfa((FFTW_REAL*)arr, (FFTW_REAL*)arr + 1, 
			       work, 2, 2 * N, N, 1, +1),
			  1.0/N,
			  compute_accuracy)
     else SKIP_BENCHMARK("Temperton only handles N = 2^m 3^n 5^q");
}

void do_temperton_f_fft(int rank, int *n, int *n_rev, int N, 
		   short is_power_of_two,
		   FFTW_COMPLEX *arr, FFTW_COMPLEX *work,
		   int size_arr, int size_work,
		   short compute_accuracy, factor_type allowed_factors)
{
#ifdef USE_FORTRAN
     if (sizeof(FFTW_REAL) == sizeof(double)) {
	  int np = 0, two = 2, twoN = 2*N, one = 1, is1=-1,is2=+1;
	  
	  if (rank != 1) return;
	  if (N != 0) {
	       /* insure that array size is power of 2, 3, & 5 only: */
	       if (check_prime_factors(N,5)) {
		    np = 0;
		    FORTRANIZE(gpfafsetgpfa,GPFAFSETGPFA)(work,&N);
	       }
	       else
		    np = 1;
	  }
	  if (np == 0)
	       DO_BENCHMARK_ND("Temperton",
			       1, n, N, arr, arr,
			       FORTRANIZE(gpfafgpfa,GPFAFGPFA)
			           ((FFTW_REAL*)arr, (FFTW_REAL*)arr + 1, 
				    work, &two, &twoN, &N, &one, &is1),
			       1.0, -1, 1,
			       FORTRANIZE(gpfafgpfa,GPFAFGPFA)
			           ((FFTW_REAL*)arr, (FFTW_REAL*)arr + 1, 
				    work, &two, &twoN, &N, &one, &is2),
			       1.0/N,
			       compute_accuracy)
	  else SKIP_BENCHMARK("Temperton only handles N = 2^m 3^n 5^q");
     }
#endif
}

void do_green_fft(int rank, int *n, int *n_rev, int N, 
		short is_power_of_two,
                FFTW_COMPLEX *arr, FFTW_COMPLEX *work,
                int size_arr, int size_work,
                short compute_accuracy, factor_type allowed_factors)
{
     if (rank != 1) return;
     if (is_power_of_two) {
	  int M = 0;
	  
	  if (N > 0) {
	       if (1 == green_FFTInit(&M,N,(FFTW_REAL*) work))
		    M = -1;
	  }
	  
	  if (M >= 0)
	       DO_BENCHMARK_ND("Green",
			       1, n, N, arr, arr,
			       green_ffts((FFTW_REAL*) arr, M, 1, 
					  (FFTW_REAL*) work),
			       1.0, -1, 1,
			       green_iffts((FFTW_REAL*) arr, M, 1, 
					   (FFTW_REAL*) work),
			       1.0,
			       compute_accuracy)
	  else
	       SKIP_BENCHMARK("Green can't handle this size.");
     }
     else if (allowed_factors != NON_POWERS_OF_TWO_ONLY)
	  SKIP_BENCHMARK("not a power of 2");
}

void do_gsl_fft(int rank, int *n, int *n_rev, int N, 
		   short is_power_of_two,
		   FFTW_COMPLEX *arr, FFTW_COMPLEX *work,
		   int size_arr, int size_work,
		   short compute_accuracy, factor_type allowed_factors)
{
     gsl_fft_complex_wavetable wt;
     int status = 0;

     if (rank != 1) return;

     /******* Mixed-radix routine ********/

     if (N != 0) {
	  /* Initialize the wavetable to point to our work array,
	     rather than reallocating.  Yes, this breaks the abstraction,
	     but it saves us memory. */
	  
	  wt.scratch = (complex*)work;
	  wt.trig = (complex*)work + N;
	  
	  /* Now, initialize: */
	  status = gsl_fft_complex_init(N,&wt);
     }
     
     if (status == 0) {
	  DO_BENCHMARK_ND("GSL",
			  1, n, N, arr, arr,
			  gsl_fft_complex_forward((complex*)arr,N,&wt), 
			  1.0, -1, 1,
			  gsl_fft_complex_backward((complex*)arr,N,&wt),
			  1.0/N,
			  compute_accuracy);
     }
     else if (N != 0)
	  SKIP_BENCHMARK("Error initializing GSL wavetable!");

     /****** Radix 2 DIT and DIF routines ********/

     if (is_power_of_two) {
	  DO_BENCHMARK_ND("GSL DIT",
			  1, n, N, arr, arr,
			  gsl_fft_complex_radix2_forward((complex*)arr,N), 
			  1.0, -1, 1,
			  gsl_fft_complex_radix2_backward((complex*)arr,N),
			  1.0/N,
			  compute_accuracy);
	  DO_BENCHMARK_ND("GSL DIF",
			  1, n, N, arr, arr,
			  gsl_fft_complex_radix2_dif_forward((complex*)arr,N),
			  1.0, -1, 1,
			  gsl_fft_complex_radix2_dif_backward((complex*)arr,N),
			  1.0/N,
			  compute_accuracy);
     }
     else if (allowed_factors != NON_POWERS_OF_TWO_ONLY) {
          SKIP_BENCHMARK("not a power of 2"); /* DIT */
          SKIP_BENCHMARK("not a power of 2"); /* DIF */
     }

}

void do_krukar_fft(int rank, int *n, int *n_rev, int N, 
		   short is_power_of_two,
		   FFTW_COMPLEX *arr, FFTW_COMPLEX *work,
		   int size_arr, int size_work,
		   short compute_accuracy, factor_type allowed_factors)
{
     if (rank != 1) return;
     if (is_power_of_two && N <= 4096) {
	  DO_BENCHMARK_ND("Krukar",
			  1, n, N, arr, arr,
			  fft_in_C(arr, N), 1.0, -1, 1,
			  ifft_in_C(arr, N), 1.0,
			  compute_accuracy);
     }
     else if (is_power_of_two && N > 4096)
	  SKIP_BENCHMARK("Krukar can't handle N > 4096")
     else if (allowed_factors != NON_POWERS_OF_TWO_ONLY)
	  SKIP_BENCHMARK("not a power of 2")
}

void do_mayer_fft(int rank, int *n, int *n_rev, int N, 
		short is_power_of_two,
                FFTW_COMPLEX *arr, FFTW_COMPLEX *work,
                int size_arr, int size_work,
                short compute_accuracy, factor_type allowed_factors)
{
     if (rank != 1) return;

     if (is_power_of_two && (N == 0 || N > 2) && N <= (1<<19))
	  DO_BENCHMARK_ND("Mayer (Buneman)",
                          1, n, N, arr, arr,
                          mayer_fft_1(N, (FFTW_REAL*)arr, (FFTW_REAL*)arr + N),
                          1.0, -1, 0,
                          mayer_ifft_1(N, (FFTW_REAL*)arr,(FFTW_REAL*)arr + N),
                          1.0,
                          compute_accuracy)
     else if (allowed_factors != NON_POWERS_OF_TWO_ONLY) {
	  if (!is_power_of_two)
	       SKIP_BENCHMARK("not a power of 2")
	  else if (N <= 2)
	       SKIP_BENCHMARK("Mayer can't handle N <= 2")
	  else
	       SKIP_BENCHMARK("Mayer (Buneman) can't handle N > 2^19")
     }

     if (is_power_of_two && (N == 0 || N > 2))
	  DO_BENCHMARK_ND("Mayer (simple)",
                          1, n, N, arr, arr,
                          mayer_fft_5(N, (FFTW_REAL*)arr, (FFTW_REAL*)arr + N),
                          1.0, -1, -1,
                          mayer_ifft_5(N, (FFTW_REAL*)arr,(FFTW_REAL*)arr + N),
                          1.0,
                          compute_accuracy)
     else if (allowed_factors != NON_POWERS_OF_TWO_ONLY) {
	  if (!is_power_of_two)
	       SKIP_BENCHMARK("not a power of 2")
	  else
	       SKIP_BENCHMARK("Mayer can't handle N <= 2")
     }

     if (is_power_of_two && (N == 0 || N > 2)) {
	  extern FFTW_REAL *mayer_hugetrigtab;
	  mayer_hugetrigtab = (FFTW_REAL*)work;
	  DO_BENCHMARK_ND("Mayer (lookup)",
                          1, n, N, arr, arr,
                          mayer_fft_7(N, (FFTW_REAL*)arr, (FFTW_REAL*)arr + N),
                          1.0, -1, 0,
                          mayer_ifft_7(N, (FFTW_REAL*)arr,(FFTW_REAL*)arr + N),
                          1.0,
                          compute_accuracy);
     }
     else if (allowed_factors != NON_POWERS_OF_TWO_ONLY) {
	  if (!is_power_of_two)
	       SKIP_BENCHMARK("not a power of 2")
	  else
	       SKIP_BENCHMARK("Mayer can't handle N <= 2")
     }
}

void do_monro_fft(int rank, int *n, int *n_rev, int N, 
		  short is_power_of_two,
		  FFTW_COMPLEX *arr, FFTW_COMPLEX *work,
		  int size_arr, int size_work,
		  short compute_accuracy, factor_type allowed_factors)
{
#ifdef USE_FORTRAN
if (sizeof(FFTW_REAL) == sizeof(double)) {
     int mN = -N;
     if (rank != 1) return;
     if (is_power_of_two && (N == 0 || N > 2) && N <= 1048576)
	  DO_BENCHMARK_ND("Monro",
                          1, n, N, arr, arr,
			  FORTRANIZE(monrofastf,MONROFASTF)
			  ((FFTW_REAL*)arr, (FFTW_REAL*)arr + N,&N),
                          1.0*N, -1, 0,
			  FORTRANIZE(monrofastf,MONROFASTF)
			  ((FFTW_REAL*)arr, (FFTW_REAL*)arr + N,&mN),
                          1.0,
                          compute_accuracy)
     else if (allowed_factors != NON_POWERS_OF_TWO_ONLY) {
	  if (!is_power_of_two)
	       SKIP_BENCHMARK("not a power of 2")
	  else if (N <= 2)
	       SKIP_BENCHMARK("Monro can't handle N <= 2")
	  else
	       SKIP_BENCHMARK("Monro can't handle N > 2^20")
     }
}
#endif
}

void do_napack_fft(int rank, int *n, int *n_rev, int N, 
		   short is_power_of_two,
		   FFTW_COMPLEX *arr, FFTW_COMPLEX *work,
		   int size_arr, int size_work,
		   short compute_accuracy, factor_type allowed_factors)
{
     if (rank != 1) return;
#if defined(linux) ||defined(__linux__) || defined(__LINUX__)
     if (is_power_of_two)  /* NAPACK has a bug for N=9 under linux gcc */
#endif
	  DO_BENCHMARK_ND("NAPACK (f2c)",
			  1, n, N, arr, arr,
			  napack_fft(arr, &N, work),
			  1.0, +1, 1,
			  0,
			  1.0/N,
			  -compute_accuracy)
#if defined(linux) ||defined(__linux__) || defined(__LINUX__)
     else if (allowed_factors != NON_POWERS_OF_TWO_ONLY)
	  SKIP_BENCHMARK("NAPACK has bug for non-powers"
			 " of 2 under linux gcc");
#endif
}

void do_nielsen_fft(int rank, int *n, int *n_rev, int N, 
		    short is_power_of_two,
		    FFTW_COMPLEX *arr, FFTW_COMPLEX *work,
		    int size_arr, int size_work,
		    short compute_accuracy, factor_type allowed_factors)
{
#ifdef USE_NIELSEN
     if (rank != 1) return;
     DO_BENCHMARK_ND("Nielsen",
		     1, n, N, arr, work,
		     mixfft(N, (FFTW_REAL*)arr, ((FFTW_REAL*)arr) + N,
			    (FFTW_REAL*)work, ((FFTW_REAL*)work) + N),
		     1.0, -1, 0,
		     0, 1.0,
		     -compute_accuracy);
#endif
}

void do_nrc_1d_fft(int rank, int *n, int *n_rev, int N, 
		   short is_power_of_two,
		   FFTW_COMPLEX *arr, FFTW_COMPLEX *work,
		   int size_arr, int size_work,
		   short compute_accuracy, factor_type allowed_factors)
{
#ifdef USE_NRC
     if (rank != 1) return;
     if (is_power_of_two) {
	  FFTW_REAL *arr2 = ((FFTW_REAL*)arr) - 1; /* NRC is 1-based! */
	  DO_BENCHMARK_ND("NR (C)",
			  1, n, N, arr, arr,
			  nrc_four1(arr2, N, -1),
			  1.0, -1, 1,
			  nrc_four1(arr2, N, +1),
			  1.0/N,
			  compute_accuracy);     	
     }
     else if (allowed_factors != NON_POWERS_OF_TWO_ONLY)
	  SKIP_BENCHMARK("not a power of 2");
#endif
}

void do_nrf_1d_fft(int rank, int *n, int *n_rev, int N, 
		   short is_power_of_two,
		   FFTW_COMPLEX *arr, FFTW_COMPLEX *work,
		   int size_arr, int size_work,
		   short compute_accuracy, factor_type allowed_factors)
{
#ifdef USE_NRF
#ifdef USE_FORTRAN
     if (rank != 1) return;
     if (is_power_of_two) {
	  int n1=N, is1=-1,is2=+1;
	  DO_BENCHMARK_ND("NR (F)",
			  1, n, N, arr, arr,
			  FORTRANIZE(nrffour1,NRFFOUR1)(arr, &n1, &is1),
			  1.0, -1, 1,
			  FORTRANIZE(nrffour1,NRFFOUR1)(arr, &n1, &is2),
			  1.0/N,
			  compute_accuracy);     	
     }
     else if (allowed_factors != NON_POWERS_OF_TWO_ONLY)
	  SKIP_BENCHMARK("not a power of 2");
#endif
#endif
}

void do_ooura_fft(int rank, int *n, int *n_rev, int N, 
		  short is_power_of_two,
		  FFTW_COMPLEX *arr, FFTW_COMPLEX *work,
		  int size_arr, int size_work,
		  short compute_accuracy, factor_type allowed_factors)
{
     if (rank != 1) return;
     if (is_power_of_two) {
	  int *ip, len_ip, n2=2*N, is1=-1, is2=+1;
	  
	  if (N != 0) {
	       if (N % 4 == 0)
		    len_ip = sqrt(N) + 3;
	       else
		    len_ip = sqrt(N/2) + 3;
	       ip = fftw_malloc(sizeof(int)*len_ip);
	       if (!ip) {
		    printf("\nERROR!  Out of memory!\n");
		    exit(1);
	       }
	       ip[0] = 0;
	       ooura_c_cdft(n2, +1, arr, ip, work);
	  }
	  DO_BENCHMARK_ND("Ooura (C)",
			  1, n, N, arr, arr,
			  ooura_c_cdft(n2, +1, arr, ip, work),
			  1.0, +1, 1,
			  ooura_c_cdft(n2, -1, arr, ip, work),
			  1.0/N,
			  compute_accuracy);
#ifdef USE_FORTRAN
	  if (N != 0) {
	       ip[0] = 0;
	       FORTRANIZE(oourafcdft,OOURAFCDFT)(&n2, &is2, arr, 
						     ip, work);
	  }
	  DO_BENCHMARK_ND("Ooura (F)",
			  1, n, N, arr, arr,
			  FORTRANIZE(oourafcdft,
				     OOURAFCDFT)(&n2, &is2, arr, 
						   ip, work),
			  1.0, +1, 1,
			  FORTRANIZE(oourafcdft,
				     OOURAFCDFT)(&n2, &is1, arr, 
						   ip, work),
			  1.0/N,
			  compute_accuracy);
#endif
	  if (N != 0)
	       fftw_free(ip);
     }
     else if (allowed_factors != NON_POWERS_OF_TWO_ONLY) {
	  SKIP_BENCHMARK("not a power of 2");
	  #ifdef USE_FORTRAN
	  SKIP_BENCHMARK("not a power of 2");
          #endif
     }
}

void do_qft_fft(int rank, int *n, int *n_rev, int N,
		short is_power_of_two,
		FFTW_COMPLEX *arr, FFTW_COMPLEX *work,
		int size_arr, int size_work,
		short compute_accuracy, factor_type allowed_factors)
{
#ifdef USE_QFT
     if (rank != 1) return;
     if (is_power_of_two) {
	  if (N == 0 || N >= 16) {
	       if (N != 0)
		    qftinit(N,(void *) (work + N));
	       
	       DO_BENCHMARK_ND("QFT",
			       1, n, N, arr, work,
			       cfqft((FFTW_REAL*)work,
				     N + (FFTW_REAL*)work,
				     (FFTW_REAL*)arr,
				     N + (FFTW_REAL*)arr,
				     N,N),
			       1.0, -1, 0,
			       0,
			       1.0/N,
			       -compute_accuracy);
	  }
	  else
	       SKIP_BENCHMARK("QFT requires N >= 16");
     }
     else if (allowed_factors != NON_POWERS_OF_TWO_ONLY)
          SKIP_BENCHMARK("not a power of 2");
#endif
}

void do_ransom_fft(int rank, int *n, int *n_rev, int N,
		   short is_power_of_two,
		   FFTW_COMPLEX *arr, FFTW_COMPLEX *work,
		   int size_arr, int size_work,
		   short compute_accuracy, factor_type allowed_factors)
{
     if (rank != 1) return;
     if (is_power_of_two) {
	  if (N != 2)
	       DO_BENCHMARK_ND("Ransom",
			       1, n, N, arr, arr,
			       tablesixstepfft(arr,N,-1),
			       1.0, -1, 1,
			       tablesixstepfft(arr,N,+1),
			       1.0/N,
			       compute_accuracy)
	  else
	       SKIP_BENCHMARK("Ransom doesn't work for N=2")
     }
     else if (allowed_factors != NON_POWERS_OF_TWO_ONLY)
          SKIP_BENCHMARK("not a power of 2");
}

void do_sciport_fft(int rank, int *n, int *n_rev, int N, 
		    short is_power_of_two,
		    FFTW_COMPLEX *arr, FFTW_COMPLEX *work,
		    int size_arr, int size_work,
		    short compute_accuracy, factor_type allowed_factors)
{
#ifdef USE_FORTRAN
     typedef void sp_fft_proc(int *init, int *is, int *n,
			      FFTW_COMPLEX *x, FFTW_COMPLEX *work,
			      FFTW_COMPLEX *y);
     typedef void (*sp_fft_proc_ptr)(int *init, int *is, int *n,
				     FFTW_COMPLEX *x, FFTW_COMPLEX *work,
				     FFTW_COMPLEX *y);
     sp_fft_proc_ptr sp_fft;
     extern sp_fft_proc
	  FORTRANIZE(spscfft2,SPSCFFT2), FORTRANIZE(spdcfft2,SPDCFFT2);

     if (rank != 1) return;

     if (sizeof(FFTW_REAL) == sizeof(double))
	  sp_fft = FORTRANIZE(spdcfft2,SPDCFFT2);
     else
	  sp_fft = FORTRANIZE(spscfft2,SPSCFFT2);

     if (is_power_of_two && (N == 0 || N >= 4)) {
	  int init = 1, is1 = -1, is2 = +1;

	  if (N != 0) 
	       sp_fft(&init,&is1,&N,arr,work,arr);

	  init = 0;

	  DO_BENCHMARK_ND("SCIPORT",
			  rank, n, N, arr, arr,
			  sp_fft(&init,&is1,&N,arr,work,arr),
			  1.0, -1, 1,
			  sp_fft(&init,&is2,&N,arr,work,arr),
			  1.0/N,
			  compute_accuracy);	  
     }
     else if (allowed_factors != NON_POWERS_OF_TWO_ONLY) {
	  if (N < 4)
	       SKIP_BENCHMARK("SCIPORT can't handle N < 4")
	  else
	       SKIP_BENCHMARK("not a power of 2");
     }
#endif
}

void do_singleton_fft(int rank, int *n, int *n_rev, int N, 
		      short is_power_of_two,
		      FFTW_COMPLEX *arr, FFTW_COMPLEX *work,
		      int size_arr, int size_work,
		      short compute_accuracy, factor_type allowed_factors)
{
#ifdef USE_FORTRAN
     if (rank != 1) return;
     if (sizeof(FFTW_REAL) == sizeof(double)) {
	  int n1 = N, n2 = N, n3 = N, is1 = -2, is2 = +2;
	  if (check_prime_factors(N,23)) {
	       DO_BENCHMARK_ND("Singleton",
			       1, n, N, arr, arr,
			       FORTRANIZE(singletonfft,
					  SINGLETONFFT)((FFTW_REAL*)arr,
							(FFTW_REAL*)arr + 1, 
							&N, &N, &N, &is1),
			       1.0, -1, 1,
			       FORTRANIZE(singletonfft,
					  SINGLETONFFT)((FFTW_REAL*)arr,
							(FFTW_REAL*)arr + 1, 
							&N, &N, &N, &is2),
			       1.0/N,
			       compute_accuracy);
	  }
	  else
	       SKIP_BENCHMARK("Singleton can't handle prime factors > 23")
     }
#endif
}

void do_singleton_f2c_fft(int rank, int *n, int *n_rev, int N, 
		short is_power_of_two,
                FFTW_COMPLEX *arr, FFTW_COMPLEX *work,
                int size_arr, int size_work,
                short compute_accuracy, factor_type allowed_factors)
{
     if (rank != 1) return;
	  if (check_prime_factors(N,23)) {
	       DO_BENCHMARK_ND("Singleton (f2c)",
			       1, n, N, arr, arr,
			       go_fft((FFTW_REAL*)arr, (FFTW_REAL*)arr + 1, 
				      N, N, N, -2),
			       1.0, -1, 1,
			       go_fft((FFTW_REAL*)arr, (FFTW_REAL*)arr + 1, 
				      N, N, N, +2),
			       1.0/N,
			       compute_accuracy);     
	  }
	  else
	       SKIP_BENCHMARK("Singleton can't handle prime factors > 23");
}

void do_sorensen_ffts(int rank, int *n, int *n_rev, int N,
		      short is_power_of_two,
		      FFTW_COMPLEX *arr, FFTW_COMPLEX *work,
		      int size_arr, int size_work,
		      short compute_accuracy, factor_type allowed_factors)
{
#ifdef USE_FORTRAN
if (sizeof(FFTW_REAL) == sizeof(double)) {
     if (rank != 1) return;
     if (is_power_of_two) {
	  int m = 0;
	  int *itab = 0;

	  if (N != 0) {
	       int len_itab;

	       while ((1 << m) < N) ++m;

	       len_itab = sqrt(2.0*N) + 1.0;
	       itab = fftw_malloc(sizeof(int) * len_itab);
	       if (!itab) {
		    printf("\nERROR! Out of memory for Sorensen!\n");
		    exit(1);
	       }
	  }

#define USE_SORENSEN_NOTAB 0 /* Sorensen (no table) is buggy...don't use */

#if USE_SORENSEN_NOTAB
	  if (N <= 2048)
	       DO_BENCHMARK_ND("Sorensen (no table)",
			       1, n, N, arr, arr,
			       FORTRANIZE(cfftsr,CFFTSR)((FFTW_REAL*)arr,
							 (FFTW_REAL*)arr + N,
							 &m,itab),
			       1.0, -1, 0,
			       0,
			       1.0/N,
			       -compute_accuracy)
	  else
	       SKIP_BENCHMARK("Sorensen has a bug for N > 2048")
#endif

	  {
	       int len;
	       FFTW_REAL *w = (FFTW_REAL*)work;

	       if (N != 0) {
		    len = (N + 7) / 8;

		    FORTRANIZE(tinit,TINIT)(&m,w,w+len,
					    w+2*len,w+3*len,itab);
	       }

	       DO_BENCHMARK_ND("Sorensen",
			       1, n, N, arr, arr,
			       FORTRANIZE(ctfftsr,CTFFTSR)((FFTW_REAL*)arr,
							   (FFTW_REAL*)arr + N,
							   &m,w,w+len,
							   w+2*len,w+3*len,
							   itab),
			       1.0, -1, 0,
			       0,
			       1.0/N,
			       -compute_accuracy);
	  }

	  if (N != 0)
	       fftw_free(itab);
     }
     else if (allowed_factors != NON_POWERS_OF_TWO_ONLY) {
#if USE_SORENSEN_NOTAB
          SKIP_BENCHMARK("not a power of 2"); /* Sorensen (no table) */
#endif
          SKIP_BENCHMARK("not a power of 2"); /* Sorensen (table) */
     }
}
#endif
}

void do_sorensen_dit_fft(int rank, int *n, int *n_rev, int N, 
			 short is_power_of_two,
			 FFTW_COMPLEX *arr, FFTW_COMPLEX *work,
			 int size_arr, int size_work,
			 short compute_accuracy, factor_type allowed_factors)
{
#ifdef USE_FORTRAN
if (sizeof(FFTW_REAL) == sizeof(double)) {
     if (rank != 1) return;
     if (is_power_of_two) {
	  int n1 = N, m = 0, is1 = -1, is2 = +1;
	  FFTW_REAL  *u = (FFTW_REAL *) (work + N);
	  
	  if (N != 0) {
	       while (n1 % 2 == 0) {
		    ++m;
		    n1 /= 2;
	       }
	  }
	  DO_BENCHMARK_ND("Sorensen DIT",
			  rank, n, N, arr, arr,
			  FORTRANIZE(sfftfu,SFFTFU)
			       ((FFTW_REAL*)arr,(FFTW_REAL*)arr + N,
				&N,&m,&is2),
			  1.0, -1, 0,
			  FORTRANIZE(sfftfu,SFFTFU)
			       ((FFTW_REAL*)arr,(FFTW_REAL*)arr + N,
				&N,&m,&is1),
			  1.0,
			  compute_accuracy);	  
     }
     else if (allowed_factors != NON_POWERS_OF_TWO_ONLY)
	  SKIP_BENCHMARK("not a power of 2");
}
#endif
}

void do_valkenburg_fft(int rank, int *n, int *n_rev, int N, 
		       short is_power_of_two,
		       FFTW_COMPLEX *arr, FFTW_COMPLEX *work,
		       int size_arr, int size_work,
		       short compute_accuracy, factor_type allowed_factors)
{
     extern void W_deallocate(void);
     if (rank != 1) return;
     DO_BENCHMARK_ND("Valkenburg",
		     1, n, N, arr, work,
		     rft(arr, N, work),
		     1.0, +1, 1,
		     fft(work, N, arr),
		     1.0,
		     compute_accuracy);
     W_deallocate();
}

void do_scilib_1d_fft(int rank, int *n, int *n_rev, int N, 
		      short is_power_of_two,
		      FFTW_COMPLEX *arr, FFTW_COMPLEX *work,
		      int size_arr, int size_work,
		      short compute_accuracy, factor_type allowed_factors)
{
#ifdef USE_SCILIB
     if (rank != 1) return;
     if (sizeof(FFTW_REAL) == sizeof(double)) {
	  int isign, isign2=+1, n1, isys[1] = {0};
	  FFTW_REAL scale = 1.0;
	  FFTW_REAL *work2 = ((FFTW_REAL*)work) + size_work;
	  
	  if (N != 0) {
	       n1 = N;
	       isign = 0;
	       CCFFT(&isign, &n1, &scale, arr, arr, 
		     work2, (FFTW_REAL*)work, 
		     isys);
	       isign = -1;
	  }
	  DO_BENCHMARK_ND("SCILIB",
			  1, n, N, arr, arr,
			  CCFFT(&isign, &n1, &scale, arr, arr,
				work2,
				(FFTW_REAL*)work,
				isys),
			  1.0, -1, 1,
			  CCFFT(&isign2, &n1, &scale, arr, arr,
				work2,
				(FFTW_REAL*)work,
				isys),
			       1.0/N,
			  compute_accuracy);
     }
#endif
}

void do_essl_1d_fft(int rank, int *n, int *n_rev, int N, 
		    short is_power_of_two,
		    FFTW_COMPLEX *arr, FFTW_COMPLEX *work,
		    int size_arr, int size_work,
		    short compute_accuracy, factor_type allowed_factors)
{
#ifdef USE_ESSL
     int n2;
	
     if (rank != 1) return;
     if (N > 0) {
	  n2 = N;
	  while (n2 % 2 == 0) n2 /= 2;
     }
     if (N == 0 || (N >= 6 && N % 2 == 0 && n2 < 10)) {
	  FFTW_REAL scale = 1.0;
	  essl_fft_proc *ESSL_FFT;
	  int n1;
	  int isign, d1, d2, d3, d4, d5, naux1, naux2, init;
	  FFTW_REAL *work2 = ((FFTW_REAL*)work) + size_work;
	  
	  n1 = N;
	  d1 = 1;
	  d2 = 0;
	  d3 = 1;
	  d4 = 0;
	  d5 = 1;
	  isign = -1;
	  naux1 = naux2 = size_work;
	  
	  if (sizeof(double) == sizeof(FFTW_REAL))
	       ESSL_FFT = dcft;
	  else
	       ESSL_FFT = scft;
	  
	  if (N != 0) {
	       init = 1;
	       ESSL_FFT(&init, arr, &d1, &d2, arr, 
			&d3, &d4, &n1, &d5, &isign, 
			&scale, (FFTW_REAL*)work, 
			&naux1, work2, &naux2);
	  }
	  init = 0;
	  
	  DO_BENCHMARK_ND("ESSL",
			  1, n, N, arr, arr,
			  ESSL_FFT(&init, arr, &d1, &d2, arr,
				   &d3, &d4, &n1, &d5, &isign,
				   &scale, (FFTW_REAL*)work, &naux1,
				   work2, &naux2),
			  1.0, +1, 1,
			  (isign = 1, init = 1,
			   ESSL_FFT(&init, arr, &d1, &d2, arr,
				    &d3, &d4, &n1, &d5, &isign,
				    &scale, (FFTW_REAL*)work, &naux1,
					      work2, &naux2),
			   init = 0,
			   ESSL_FFT(&init, arr, &d1, &d2, arr,
				    &d3, &d4, &n1, &d5, &isign,
				    &scale, (FFTW_REAL*)work, &naux1,
				    work2, &naux2)),
			  1.0/N,
			  compute_accuracy);
	  
     }
     else
	  SKIP_BENCHMARK("ESSL only works for N = x * 2^m, "
			 "where x < 10 and m >= 1");
#endif
}

void do_sunperf_fft(int rank, int *n, int *n_rev, int N, 
		    short is_power_of_two,
		    FFTW_COMPLEX *arr, FFTW_COMPLEX *work,
		    int size_arr, int size_work,
		    short compute_accuracy, factor_type allowed_factors)
{
#ifdef USE_SUNPERF
     typedef void ffti_proc(int,FFTW_COMPLEX *);
     typedef void fftfb_proc(int,FFTW_COMPLEX *,FFTW_COMPLEX *);
     typedef void (*ffti_proc_ptr)(int,FFTW_COMPLEX *);
     typedef void (*fftfb_proc_ptr)(int,FFTW_COMPLEX *,
				    FFTW_COMPLEX *);
     ffti_proc_ptr sunperf_ffti;
     fftfb_proc_ptr sunperf_fftf, sunperf_fftb;
     extern ffti_proc zffti,cffti;
     extern fftfb_proc zfftf, zfftb, cfftf, cfftb;
 
     if (rank != 1) return;
    
     if (sizeof(double) == sizeof(FFTW_REAL)) {
	  sunperf_ffti = zffti;
	  sunperf_fftf = zfftf;
	  sunperf_fftb = zfftb;
     }
     else {
	  sunperf_ffti = cffti;
	  sunperf_fftf = cfftf;
	  sunperf_fftb = cfftb;
     }
     
     if (N != 0) sunperf_ffti(N, work);
     DO_BENCHMARK_ND("SUNPERF",
		     1, n, N, arr, arr,
		     sunperf_fftf(N, arr, work),
		     1.0, -1, 1,
		     sunperf_fftb(N, arr, work),
		     1.0/N,
		     compute_accuracy);     	  
#endif
}


void do_sgimath_fft(int rank, int *n, int *n_rev, int N, 
		    short is_power_of_two,
		    FFTW_COMPLEX *arr, FFTW_COMPLEX *work,
		    int size_arr, int size_work,
		    short compute_accuracy, factor_type allowed_factors)
{
#ifdef USE_SGIMATH
     if (rank != 1) return;
    
     if (sizeof(double) == sizeof(FFTW_REAL)) {
	  FFTW_COMPLEX *zfft1di( int n, FFTW_COMPLEX *save);
	  int zfft1d(int sign, int n, FFTW_COMPLEX *array, 
		     int inc, FFTW_COMPLEX *save);
	  if (N != 0) zfft1di(N, work);
	  DO_BENCHMARK_ND("SGIMATH",
			  1, n, N, arr, arr,
			  zfft1d(-1, N, arr, 1, work),
			  1.0, -1, 1,
			  zfft1d(1, N, arr, 1, work),
			  1.0/N,
			  compute_accuracy);
     } else {
	  FFTW_COMPLEX *cfft1di(int n, FFTW_COMPLEX *save);
	  int cfft1d(int sign, int n, FFTW_COMPLEX *array, 
		     int inc, FFTW_COMPLEX *save);
	  if (N != 0) cfft1di(N, work);
	  DO_BENCHMARK_ND("SGIMATH",
			  1, n, N, arr, arr,
			  cfft1d(-1, N, arr, 1, work),
			  1.0, -1, 1,
			  cfft1d(1, N, arr, 1, work),
			  1.0/N,
			  compute_accuracy);
     }
#endif
}

void do_scsl_fft(int rank, int *n, int *n_rev, int N, 
		 short is_power_of_two,
		 FFTW_COMPLEX *arr, FFTW_COMPLEX *work,
		 int size_arr, int size_work,
		 short compute_accuracy, factor_type allowed_factors)
{
#ifdef USE_SCSL
     typedef void scsl_fft_proc(int *isign, int *n, FFTW_REAL *scale,
			   FFTW_COMPLEX *x, FFTW_COMPLEX *y,
			   FFTW_REAL *table, FFTW_REAL *work, 
			   int *isys);     
     typedef void (*scsl_fft_proc_ptr)(int *isign, int *n, FFTW_REAL *scale,
				       FFTW_COMPLEX *x, FFTW_COMPLEX *y,
				       FFTW_REAL *table, FFTW_REAL *work, 
				       int *isys);
     extern scsl_fft_proc
	  FORTRANIZE(ccfft,CCFFT), FORTRANIZE(zzfft,ZZFFT);
     scsl_fft_proc_ptr scsl_fft;
     int isign=-1, isign2=+1, n1 = N, isys = 0;
     FFTW_REAL scale = 1.0;
     FFTW_REAL *table, *work2;

     if (rank != 1) return;

     table = (FFTW_REAL*)work;
     work2 = table + 2*N + 30;

     if (sizeof(FFTW_REAL) == sizeof(double))
	  scsl_fft = FORTRANIZE(zzfft,ZZFFT);
     else
	  scsl_fft = FORTRANIZE(ccfft,CCFFT);

     if (N != 0) {
	  n1 = N;
	  isign = 0;
	  scsl_fft(&isign, &n1, &scale, arr, arr, 
		   table, work2, &isys);
	  isign = -1;
     }
     DO_BENCHMARK_ND("SCSL",
		     1, n, N, arr, arr,
		     scsl_fft(&isign, &n1, &scale, arr, arr, 
			      table, work2, &isys),
		     1.0, -1, 1,
		     scsl_fft(&isign2, &n1, &scale, arr, arr, 
			      table, work2, &isys),
		     1.0/N,
		     compute_accuracy);
#endif
}

void do_imsl_fft(int rank, int *n, int *n_rev, int N, 
		 short is_power_of_two,
		 FFTW_COMPLEX *arr, FFTW_COMPLEX *work,
		 int size_arr, int size_work,
		 short compute_accuracy, factor_type allowed_factors)
{
#ifdef USE_IMSL
     typedef void ffti_proc(int *,FFTW_COMPLEX *);
     typedef void fftfb_proc(int *,FFTW_COMPLEX *,FFTW_COMPLEX *,
			     FFTW_COMPLEX *,FFTW_COMPLEX *);
     typedef void (*ffti_proc_ptr)(int *,FFTW_COMPLEX *);
     typedef void (*fftfb_proc_ptr)(int *,FFTW_COMPLEX *,FFTW_COMPLEX *,
				    FFTW_COMPLEX *,FFTW_COMPLEX *);
     ffti_proc_ptr imsl_ffti;
     fftfb_proc_ptr imsl_fftf, imsl_fftb;
     extern ffti_proc FORTRANIZE(fftci,FFTCI), FORTRANIZE(dfftci,DFFTCI);
     extern fftfb_proc 
	  FORTRANIZE(f2tcf,F2TCF), FORTRANIZE(df2tcf,DF2TCF),
	  FORTRANIZE(f2tcb,F2TCB), FORTRANIZE(df2tcb,DF2TCB);

     if (rank != 1) return;
    
     if (sizeof(float) == sizeof(FFTW_REAL)) {
	  imsl_ffti = FORTRANIZE(fftci,FFTCI);
	  imsl_fftf = FORTRANIZE(f2tcf,F2TCF);
	  imsl_fftb = FORTRANIZE(f2tcb,F2TCB);
     }
     else {
	  imsl_ffti = FORTRANIZE(dfftci,DFFTCI);
	  imsl_fftf = FORTRANIZE(df2tcf,DF2TCF);
	  imsl_fftb = FORTRANIZE(df2tcb,DF2TCB);
     }
     
     if (N != 0) imsl_ffti(&N, work + N);
     DO_BENCHMARK_ND("IMSL",
		     1, n, N, arr, arr,
		     imsl_fftf(&N, arr, arr, work + N, work),
		     1.0, -1, 1,
		     imsl_fftb(&N, arr, arr, work + N, work),
		     1.0/N,
		     compute_accuracy);     	  
#endif
}

void do_nag_fft(int rank, int *n, int *n_rev, int N, 
		 short is_power_of_two,
		 FFTW_COMPLEX *arr, FFTW_COMPLEX *work,
		 int size_arr, int size_work,
		 short compute_accuracy, factor_type allowed_factors)
{
#ifdef USE_NAG
     typedef void fftfb_proc(FFTW_REAL *r, FFTW_REAL *i, int *n,
			     FFTW_REAL *work, int *fail);
     typedef void (*fftfb_proc_ptr)(FFTW_REAL *r, FFTW_REAL *i, int *n,
				    FFTW_REAL *work, int *fail);
     fftfb_proc_ptr nag_fftf;
     int fail = 0; /* MUST be initialized to 0 for NAG */
     FFTW_REAL *re,*im;
     extern fftfb_proc 
	  FORTRANIZE(c06fcf,C06FCF), FORTRANIZE(c06fce,C06FCE);

     if (rank != 1) return;
    
     if (sizeof(double) == sizeof(FFTW_REAL)) {
	  nag_fftf = FORTRANIZE(c06fcf,C06FCF);
     }
     else {
	  nag_fftf = FORTRANIZE(c06fce,C06FCE);
     }

     re = (FFTW_REAL *)arr;
     im = re + N;

     /* Note: NAG has no initialization or inverse routines. */
     
     if (N != 0) {
	  /* Call once to check for failures: */
	  nag_fftf(re,im,&N,(FFTW_REAL*)work,&fail);
     }
     if (fail == 0) {
	  DO_BENCHMARK_ND("NAG",
			  1, n, N, arr, arr,
			  nag_fftf(re,im,&N,(FFTW_REAL*)work,&fail),
			  sqrt(N), -1, 0,
			  0,
			  1.0/N,
			  -compute_accuracy);     	  
     }
     else {
	  if (fail == 1)
	       SKIP_BENCHMARK("NAG can't handle factors > 19")
	  else if (fail == 2)
	       SKIP_BENCHMARK("NAG can't handle more than 20 prime factors")
	  else if (fail == 3) 
	       SKIP_BENCHMARK("NAG can't handle N <= 1")
	  else
	       SKIP_BENCHMARK("Unknown error in NAG")
     }
#endif
}

#ifdef USE_DXML
#include <dxmldef.h>
#endif

void do_dxml_fft(int rank, int *n, int *n_rev, int N, 
		 short is_power_of_two,
		 FFTW_COMPLEX *arr, FFTW_COMPLEX *work,
		 int size_arr, int size_work,
		 short compute_accuracy, factor_type allowed_factors)
{
#ifdef USE_DXML
     if (rank != 1) return;
    
     if (sizeof(float) == sizeof(FFTW_REAL)) {
           int is_stride_1 = 1, stride = 1;
           DXML_C_FFT_STRUCTURE fft_struct;

	   if (N != 0) cfft_init_(&N,&fft_struct,&is_stride_1);
	   
	   DO_BENCHMARK_ND("DXML",
                     1, n, N, arr, arr,
                     cfft_apply_("C","C","F",arr,arr,&fft_struct,&stride),
                     1.0, -1, 1,
                     cfft_apply_("C","C","B",arr,arr,&fft_struct,&stride),
                     1.0/N,
                     compute_accuracy);

	   if (N != 0) cfft_exit_(&fft_struct);
     }
     else if (sizeof(double) == sizeof(FFTW_REAL)) {
           int is_stride_1 = 1, stride = 1;
           DXML_Z_FFT_STRUCTURE fft_struct;

	   if (N != 0) zfft_init_(&N,&fft_struct,&is_stride_1);
	   
	   DO_BENCHMARK_ND("DXML",
                     1, n, N, arr, arr,
                     zfft_apply_("C","C","F",arr,arr,&fft_struct,&stride),
                     1.0, -1, 1,
                     zfft_apply_("C","C","B",arr,arr,&fft_struct,&stride),
                     1.0/N,
                     compute_accuracy);

	   if (N != 0) zfft_exit_(&fft_struct);
     }
#endif
}

/**************************** 3D FFTs: *******************************/

extern int maxn(int n, int nums[]);
extern int max2n(int n, int nums[]);

short do_fftwnd_fft(int rank, int *n, int *n_rev, int N, 
		    short is_power_of_two,
		    FFTW_COMPLEX *arr, FFTW_COMPLEX *work,
		    int size_arr, int size_work,
		    short compute_accuracy, factor_type allowed_factors)
{
     fftwnd_plan p, p_inv;
     int wis_flags = 0;

     #ifdef FFTW_HAS_WISDOM /* if we are using a version of FFTW with the
			       "wisdom" feature, then take advantage of it */
     wis_flags = FFTW_USE_WISDOM;
     #endif
     
     if (N != 0) {
	  p = fftwnd_create_plan(rank,n,FFTW_FORWARD,
				 FFTW_MEASURE|FFTW_IN_PLACE|wis_flags);
	  p_inv = fftwnd_create_plan(rank,n,FFTW_BACKWARD,
				     FFTW_ESTIMATE|FFTW_IN_PLACE);
	  if (!p || !p_inv) {
	       fprintf(bench_log_file,
		       "\n\nError creating FFTWND plan!\n");
	       printf("\n\nError creating FFTWND plan!\n");
	       return 1;
	  }
     }
     
     DO_BENCHMARK_ND("FFTW",
		     rank, n, N, arr, arr,
		     fftwnd(p,1,arr,1,0,arr,1,0),
		     1.0, FFTW_FORWARD, 1,
		     fftwnd(p_inv,1,arr,1,0,arr,1,0), 1.0/N,
		     compute_accuracy);
     
     if (N != 0) {
	  fftwnd_destroy_plan(p);
	  fftwnd_destroy_plan(p_inv);
     }
     return 0;
}

void do_temperton_3d_fft(int rank, int *n, int *n_rev, int N, 
		    short is_power_of_two,
		    FFTW_COMPLEX *arr, FFTW_COMPLEX *work,
		    int size_arr, int size_work,
		    short compute_accuracy, factor_type allowed_factors)
{
     int okay;
     
     if (rank != 3) return;
     if (check_prime_factors(N,5) && 
	 (N == 0 || 0 == gpf3d(arr, n_rev[0], n_rev, +1)))
	  okay = 1;
     else
	  okay = 0;
     /* N.B. The +1/-1 parameter in GPF3D means the opposite
	of the usual (it is the negative of the sign of the
	exponent in the transform). */
     if (okay)
	  DO_BENCHMARK_ND("Temperton (f2c)",
			  3, n, N, arr, arr,
			  gpf3d(arr, n_rev[0], n_rev, +1),
			  1.0, -1, 1,
			  gpf3d(arr, n_rev[0], n_rev, -1),
			  1.0/N,
			  compute_accuracy)
     else if (check_prime_factors(N,5) && maxn(3,n) > 256)
	  SKIP_BENCHMARK("Temperton can't handle dimensions > 256")
     else 
	  SKIP_BENCHMARK("Temperton only handles N = 2^m 3^n 5^q");
}

void do_temperton_f_3d_fft(int rank, int *n, int *n_rev, int N, 
		      short is_power_of_two,
		      FFTW_COMPLEX *arr, FFTW_COMPLEX *work,
		      int size_arr, int size_work,
		      short compute_accuracy, factor_type allowed_factors)
{
#ifdef USE_FORTRAN
     if (sizeof(FFTW_REAL) == sizeof(double)) {
	  int np = 0, is1=+1,is2=-1;
	  
	  if (rank != 3) return;
	  if (N != 0) {
	       /* insure that array size is power of 2, 3, & 5 only: */
	       if (check_prime_factors(N,5) && maxn(3,n) <= 256) {
		    np = 0;
		    FORTRANIZE(gpfafgpf3d,GPFAFGPF3D)
			 (arr, &n_rev[0], n_rev, &is1);
	       }
	       else
		    np = 1;
	  }
	  /* N.B. The +1/-1 parameter in GPF3D means the opposite
	     of the usual (it is the negative of the sign of the
	     exponent in the transform). */
	  if (np == 0)
	       DO_BENCHMARK_ND("Temperton",
			       3, n, N, arr, arr,
			       FORTRANIZE(gpfafgpf3d,GPFAFGPF3D)
			            (arr, &n_rev[0], n_rev, &is1),
			       1.0, -1, 1,
			       FORTRANIZE(gpfafgpf3d,GPFAFGPF3D)
			            (arr, &n_rev[0], n_rev, &is2),
			       1.0/N,
			       compute_accuracy)
	  else if (np == 1 && maxn(3,n) > 256)
	       SKIP_BENCHMARK("Temperton can't handle dimensions > 256")
	  else 
	       SKIP_BENCHMARK("Temperton only handles N = 2^m 3^n 5^q");
     }
#endif
}

void do_mfft_3d_fft(int rank, int *n, int *n_rev, int N, 
		    short is_power_of_two,
		    FFTW_COMPLEX *arr, FFTW_COMPLEX *work,
		    int size_arr, int size_work,
		    short compute_accuracy, factor_type allowed_factors)
{
#ifdef USE_FORTRAN
     /* Auggh!  MFFT, because of a totally non-portable hack that
	it uses, requires that the integer and floating point types
	be the same size!  Damn these Fortran programmers! */
     if (sizeof(FFTW_REAL) == sizeof(float) &&
	 sizeof(FFTW_REAL) == sizeof(int) /* <-- obscenity */) {

	  int np = 1, is0=0,is1=-1,is2=+1;
	  int *wl=0,*wm=0,*wn=0,*iwork=0;
	  int iopt=0,iord=1,ierr;
	  
	  if (rank != 3) return;
	  if (N != 0) {
	       /* insure that array size is power of 2, 3, & 5 only: */
	       if (check_prime_factors(N,5)) {
		    wl = fftw_malloc(sizeof(int)*(4*n_rev[0]+14));
		    wm = fftw_malloc(sizeof(int)*(4*n_rev[1]+14));
		    wn = fftw_malloc(sizeof(int)*(4*n_rev[2]+14));
		    iwork = fftw_malloc(sizeof(int)*maxn(3,n_rev));
		    if (wl == 0 || wm == 0 || wn == 0 || iwork == 0) {
			 printf("\nERROR!  Out of memory for MFFT!\n");
			 exit(1);
		    }

		    /* Initialize work arrays: */
		    FORTRANIZE(mfftc3fft,MFFT_C3FFT)
			 (arr,&n_rev[0],&n_rev[0],&n_rev[1],&n_rev[2],
			  wl,wm,wn,&iopt,&is0,&iord,iwork,&ierr);

		    np = ierr;
	       }
	       else
		    np = 1;
	  }
	  if (np == 0)
	       DO_BENCHMARK_ND("MFFT",
			       3, n, N, arr, arr,
			       FORTRANIZE(mfftc3fft,MFFT_C3FFT)
			       (arr,&n_rev[0],&n_rev[0],&n_rev[1],&n_rev[2],
				wl,wm,wn,&iopt,&is1,&iord,iwork,&ierr),
			       1.0, -1, 1,
			       FORTRANIZE(mfftc3fft,MFFT_C3FFT)
			       (arr,&n_rev[0],&n_rev[0],&n_rev[1],&n_rev[2],
				wl,wm,wn,&iopt,&is2,&iord,iwork,&ierr),
			       1.0/N,
			       compute_accuracy)
	  else 
	       SKIP_BENCHMARK("MFFT only handles N = 2^m 3^n 5^q");

	  if (N == 0) {
	       if (wl) fftw_free(wl);
	       if (wm) fftw_free(wm);
	       if (wn) fftw_free(wn);
	       if (iwork) fftw_free(iwork);
	  }
     }
#endif
}

void do_harm_fft(int rank, int *n, int *n_rev, int N, 
		 short is_power_of_two,
		 FFTW_COMPLEX *arr, FFTW_COMPLEX *work,
		 int size_arr, int size_work,
		 short compute_accuracy, factor_type allowed_factors)
{
#ifdef USE_FORTRAN
if (sizeof(FFTW_REAL) == sizeof(double)) {
     if (rank != 3) return;
     if (is_power_of_two && (N == 0 || (n[0]>4 && n[1]>4 && n[2]>4))) {
	  int m[3];
	  int ifset, ifset2 = -2, iferr, *inv;
	  
	  if (N != 0) {
	       int i;
	       for (i = 0; i < 3; ++i) {
		    int lg = 0,n1 = n_rev[i];
		    while (n1 % 2 == 0) { ++lg; n1 /= 2; }
		    m[i] = lg;
	       }
	       inv = fftw_malloc(sizeof(int)*maxn(3,n));
	       if (!inv) {
		    printf("\n\nERROR!  Out of memory!\n");
		    if (bench_log_file)
			 fprintf(bench_log_file,
				 "\n\nERROR!  Out of memory!\n");
		    exit(1);
	       }
	       ifset = 0;
	       FORTRANIZE(harm,HARM)(arr, m, inv, work, &ifset, &iferr);
	       ifset = +2;
	  }
	  DO_BENCHMARK_ND("HARM",
			  3, n, N, arr, arr,
			  FORTRANIZE(harm,HARM)
			       (arr, m, inv, work, &ifset, &iferr),
			  1.0, +1, 1,
			  FORTRANIZE(harm,HARM)
			       (arr, m, inv, work, &ifset2, &iferr),
			  1.0,
			  compute_accuracy);
	  if (N != 0)
	       fftw_free(inv);
     }
     else if (is_power_of_two)
	  SKIP_BENCHMARK("all dimensions must be > 4 for HARM")
     else if (allowed_factors != NON_POWERS_OF_TWO_ONLY)
	  SKIP_BENCHMARK("not a power of 2");
}
#endif
}

void do_harm_f2c_fft(int rank, int *n, int *n_rev, int N, 
		     short is_power_of_two,
		     FFTW_COMPLEX *arr, FFTW_COMPLEX *work,
		     int size_arr, int size_work,
		     short compute_accuracy, factor_type allowed_factors)
{
     if (rank != 3) return;
     if (is_power_of_two && (N == 0 || (n[0]>4 && n[1]>4 && n[2]>4))) {
	  int m[3];
	  int ifset, ifset2 = -2, iferr, *inv;
	  
	  if (N != 0) {
	       int i;
	       for (i = 0; i < 3; ++i) {
		    int lg = 0,n1 = n_rev[i];
		    while (n1 % 2 == 0) { ++lg; n1 /= 2; }
		    m[i] = lg;
	       }
	       inv = fftw_malloc(sizeof(int)*maxn(3,n));
	       if (!inv) {
		    printf("\n\nERROR!  Out of memory!\n");
		    if (bench_log_file)
			 fprintf(bench_log_file,
				 "\n\nERROR!  Out of memory!\n");
		    exit(1);
	       }
	       ifset = 0;
	       harmd(arr, m, inv, work, &ifset, &iferr);
	       ifset = +2;
	  }
	  DO_BENCHMARK_ND("HARM (f2c)",
			  3, n, N, arr, arr,
			  harmd(arr, m, inv, work, &ifset, &iferr),
			  1.0, +1, 1,
			  harmd(arr, m, inv, work, &ifset2, &iferr),
                               1.0,
			  compute_accuracy);
	  if (N != 0)
	       fftw_free(inv);
     }
     else if (is_power_of_two)
	  SKIP_BENCHMARK("all dimensions must be > 4 for HARM")
     else if (allowed_factors != NON_POWERS_OF_TWO_ONLY)
	  SKIP_BENCHMARK("not a power of 2");
}

void do_nrc_fourn_fft(int rank, int *n, int *n_rev, int N, 
		      short is_power_of_two,
		      FFTW_COMPLEX *arr, FFTW_COMPLEX *work,
		      int size_arr, int size_work,
		      short compute_accuracy, factor_type allowed_factors)
{
#ifdef USE_NRC
     if (is_power_of_two) {
	  FFTW_REAL *arr2 = (FFTW_REAL*)arr - 1; /* 1-based! */
	  int *n1 = n - 1; /* 1-based! */
	  DO_BENCHMARK_ND("NR (C)",
			  rank, n, N, arr, arr,
			  nrc_fourn(arr2, n1, rank, -1),
			  1.0, -1, 1,
			  nrc_fourn(arr2, n1, rank, +1),
			  1.0/N,
			  compute_accuracy);
     }
     else if (allowed_factors != NON_POWERS_OF_TWO_ONLY)
	  SKIP_BENCHMARK("not a power of 2");
#endif
}

void do_nrf_fourn_fft(int rank, int *n, int *n_rev, int N, 
		      short is_power_of_two,
		      FFTW_COMPLEX *arr, FFTW_COMPLEX *work,
		      int size_arr, int size_work,
		      short compute_accuracy, factor_type allowed_factors)
{
#ifdef USE_NRF
#ifdef USE_FORTRAN
     if (is_power_of_two) {
	  int is1 = -1, is2 = +1;
	  DO_BENCHMARK_ND("NR (F)",
			  rank, n, N, arr, arr,
			  FORTRANIZE(nrffourn,NRF_FOURN)
			  (arr, n_rev, &rank, &is1),
			  1.0, -1, 1,
			  FORTRANIZE(nrffourn,NRF_FOURN)
			  (arr, n_rev, &rank, &is2),
			  1.0/N,
			  compute_accuracy);
     }
     else if (allowed_factors != NON_POWERS_OF_TWO_ONLY)
	  SKIP_BENCHMARK("not a power of 2");
#endif
#endif
}

void do_pda_fft(int rank, int *n, int *n_rev, int N, 
		    short is_power_of_two,
		    FFTW_COMPLEX *arr, FFTW_COMPLEX *work,
		    int size_arr, int size_work,
		    short compute_accuracy, factor_type allowed_factors)
{
#ifdef USE_FORTRAN
     int result;
     
     if (sizeof(FFTW_REAL) == sizeof(double)) {

     DO_BENCHMARK_ND("PDA",
		     rank, n, N, arr, arr,
		     FORTRANIZE(pdadnfftf,PDADNFFTF)(&rank, n_rev, 
						       (FFTW_REAL*)arr, 
						       (FFTW_REAL*)arr + N, 
						       work, &result),
		     sqrt(1.0*N), -1, 0,
		     FORTRANIZE(pdadnfftb,PDADNFFTB)(&rank, n_rev, 
						       (FFTW_REAL*)arr, 
						       (FFTW_REAL*)arr + N, 
						       work, &result),
		     1.0,
		     compute_accuracy);
     }
     else if (sizeof(FFTW_REAL) == sizeof(float)) {

     DO_BENCHMARK_ND("PDA",
		     rank, n, N, arr, arr,
		     FORTRANIZE(pdanfftf,PDANFFTF)(&rank, n_rev, 
						       (FFTW_REAL*)arr, 
						       (FFTW_REAL*)arr + N, 
						       work, &result),
		     sqrt(1.0*N), -1, 0,
		     FORTRANIZE(pdanfftb,PDANFFTB)(&rank, n_rev, 
						       (FFTW_REAL*)arr, 
						       (FFTW_REAL*)arr + N, 
						       work, &result),
		     1.0,
		     compute_accuracy);
     }
#endif
}

void do_pda_f2c_fft(int rank, int *n, int *n_rev, int N, 
		    short is_power_of_two,
		    FFTW_COMPLEX *arr, FFTW_COMPLEX *work,
		    int size_arr, int size_work,
		    short compute_accuracy, factor_type allowed_factors)
{
     int result;
     
     DO_BENCHMARK_ND("PDA (f2c)",
		     rank, n, N, arr, arr,
		     pda_f2c_dnfftf(&rank, n_rev, 
				(FFTW_REAL*)arr, 
				(FFTW_REAL*)arr + N, 
				work, &result),
		     sqrt(1.0*N), -1, 0,
		     pda_f2c_dnfftb(&rank, n_rev, 
				(FFTW_REAL*)arr, 
				(FFTW_REAL*)arr + N, 
				work, &result),
		     1.0,
		     compute_accuracy);
}

void do_singleton_3d_fft(int rank, int *n, int *n_rev, int N, 
			 short is_power_of_two,
			 FFTW_COMPLEX *arr, FFTW_COMPLEX *work,
			 int size_arr, int size_work,
			 short compute_accuracy, factor_type allowed_factors)
{
#ifdef USE_FORTRAN
     int N1 = N, n1, n2, n3, prod2, is1 = -2, is2 = +2;
     
     if (rank != 3) return;
     n1 = n[2];
     n2 = n[1];
     n3 = n[0];
     
     prod2 = n1 * n2;
     
     if (sizeof(FFTW_REAL) == sizeof(double))
	  if (check_prime_factors(N,23)) {
	       DO_BENCHMARK_ND("Singleton",
			  3, n, N, arr, arr,
			  {
			       FORTRANIZE(singletonfft,SINGLETONFFT)
				    ((FFTW_REAL*)arr, (FFTW_REAL*)arr+1, 
				     &N1, &n1, &n1, &is1);
			       FORTRANIZE(singletonfft,SINGLETONFFT)
				    ((FFTW_REAL*)arr, (FFTW_REAL*)arr+1, 
				     &N1, &n2, &prod2, &is1);
			       FORTRANIZE(singletonfft,SINGLETONFFT)
				    ((FFTW_REAL*)arr, (FFTW_REAL*)arr+1, 
				     &N1, &n3, &N1, &is1);
			  },
			  1.0, -1, 1,
			  {
			       FORTRANIZE(singletonfft,SINGLETONFFT)
				    ((FFTW_REAL*)arr, (FFTW_REAL*)arr+1, 
				     &N1, &n1, &n1, &is2);
			       FORTRANIZE(singletonfft,SINGLETONFFT)
				    ((FFTW_REAL*)arr, (FFTW_REAL*)arr+1, 
				     &N1, &n2, &prod2, &is2);
			       FORTRANIZE(singletonfft,SINGLETONFFT)
				    ((FFTW_REAL*)arr, (FFTW_REAL*)arr+1, 
				     &N1, &n3, &N1, &is2);
			  },
			  1.0/N,
			  compute_accuracy);     
          }
          else
               SKIP_BENCHMARK("Singleton can't handle prime factors > 23");
#endif
}


void do_singleton_3d_f2c_fft(int rank, int *n, int *n_rev, int N, 
			     short is_power_of_two,
			     FFTW_COMPLEX *arr, FFTW_COMPLEX *work,
			     int size_arr, int size_work,
			     short compute_accuracy, 
			     factor_type allowed_factors)
{
     int n1, n2, n3, prod2;
     
     if (rank != 3) return;
     n1 = n[2];
     n2 = n[1];
     n3 = n[0];
     
     prod2 = n1 * n2;
     
	  if (check_prime_factors(N,23)) {
	       DO_BENCHMARK_ND("Singleton (f2c)",
			       3, n, N, arr, arr,
		     {
			  go_fft((FFTW_REAL*)arr, (FFTW_REAL*)arr+1, 
				 N, n1, n1, -2);
			  go_fft((FFTW_REAL*)arr, (FFTW_REAL*)arr+1, 
				 N, n2, prod2, -2);
			  go_fft((FFTW_REAL*)arr, (FFTW_REAL*)arr+1, 
				 N, n3, N, -2);
		     },
		     1.0, -1, 1,
		     {
			  go_fft((FFTW_REAL*)arr, (FFTW_REAL*)arr+1, 
				 N, n1, n1, +2);
			  go_fft((FFTW_REAL*)arr, (FFTW_REAL*)arr+1, 
				 N, n2, prod2, +2);
			  go_fft((FFTW_REAL*)arr, (FFTW_REAL*)arr+1, 
				 N, n3, N, +2);
		     },
		     1.0/N,
		     compute_accuracy);     
          }
          else
               SKIP_BENCHMARK("Singleton can't handle prime factors > 23");
}

void do_essl_3d_fft(int rank, int *n, int *n_rev, int N, 
		    short is_power_of_two,
		    FFTW_COMPLEX *arr, FFTW_COMPLEX *work,
		    int size_arr, int size_work,
		    short compute_accuracy, factor_type allowed_factors)
{
#ifdef USE_ESSL
     /* we only have the single-precision version of this routine: */
     if (sizeof(FFTW_REAL) == sizeof(float)) {
	  int ok = 1;
	  if (N != 0) {
	       int i;
	       for (i = 0; i < 3; ++i) {
		    int n1 = n[i];
		    while (n1 % 2 == 0) n1 /= 2;
		    if (n1 > 9)
			 ok = 0;
		    if (n[i] % 2 != 0 || n[i] < 6)
			 ok = 0;
	       }
	  }
	  if (ok) {
	       int inc2x, inc3x, n1, n2, n3, isign, is2 = +1, naux;
	       FFTW_REAL scale = 1.0;
	       
	       naux = size_work;
	       n1 = n[2];
	       n2 = n[1];
	       n3 = n[0];
	       inc2x = n1;
	       inc3x = n1 * n2;
	       isign = -1;
	       
	       DO_BENCHMARK_ND("ESSL",
			       3, n, N, arr, arr,
			       scft3(arr, &inc2x, &inc3x, arr, 
				     &inc2x, &inc3x, 
				     &n1, &n2, &n3, &isign, 
				     &scale, work, &naux),
			       1.0, +1, 1,
			       scft3(arr, &inc2x, &inc3x, arr, 
				     &inc2x, &inc3x, 
				     &n1, &n2, &n3, &is2, 
				     &scale, work, &naux),
			       1.0/N,
			       compute_accuracy);
	  }
	  else
	       SKIP_BENCHMARK("ESSL only works for N = x * 2^m, "
			      "where x < 10 and m >= 1");
     }
#endif
}

void do_scilib_3d_fft(int rank, int *n, int *n_rev, int N, 
		      short is_power_of_two,
		      FFTW_COMPLEX *arr, FFTW_COMPLEX *work,
		      int size_arr, int size_work,
		      short compute_accuracy, factor_type allowed_factors)
{
#ifdef USE_SCILIB
     if (sizeof(FFTW_REAL) == sizeof(double)) {
	  int isign, isign2 = +1,n1, n2, n3, ldx, ldy, ldx2, ldy2;
	  FFTW_REAL scale = 1.0;
	  FFTW_REAL *cray_fft_work, *cray_fft_table;
	  int cray_fft_isys[4];
	  
	  ldx = ldy = n1 = n_rev[0];
	  ldx2 = ldy2 = n2 = n_rev[1];
	  n3 = n_rev[2];
	  
#ifdef _CRAYMPP
	  if (N != 0) {
	       int i;
	       for (i = 0; i < 3; ++i) {
		    cray_fft_isys[1+i] = !check_prime_factors(n_rev[i],5);
	       }
	       cray_fft_isys[0] = 3;
	       cray_fft_work = (FFTW_REAL*)work;
	       cray_fft_table = cray_fft_work + 2*(N);
	  }
#else
	  cray_fft_isys[0] = 1;
	  cray_fft_table = (FFTW_REAL*)work;
	  cray_fft_work = cray_fft_table + 100 + 2*(n[0]+n[1]+n[2]);
#endif
	  
	  if (N != 0) {
	       isign = 0;  /*
			    * tells FFT we are initializing */	       
	       CCFFT3D(&isign, &n1, &n2, &n3, &scale,
		       arr, &ldx, &ldx2, arr, &ldy, &ldy2,
		       cray_fft_table, cray_fft_work, cray_fft_isys);
	  }
	  isign = -1;
	  scale = 1.0;
	  
	  DO_BENCHMARK_ND("SCILIB",
			  3, n, N, arr, arr,
			  CCFFT3D(&isign, &n1, &n2, &n3, &scale,
				  arr, &ldx, &ldx2, arr, &ldy, &ldy2,
				  cray_fft_table, cray_fft_work, 
				  cray_fft_isys),
			  1.0, -1, 1,
			  CCFFT3D(&isign2, &n1, &n2, &n3, &scale,
				  arr, &ldx, &ldx2, arr, &ldy, &ldy2,
				  cray_fft_table, cray_fft_work, 
				  cray_fft_isys),
			  1.0/N,
			  compute_accuracy);
     }
#endif
}

void do_scsl_3d_fft(int rank, int *n, int *n_rev, int N, 
		      short is_power_of_two,
		      FFTW_COMPLEX *arr, FFTW_COMPLEX *work,
		      int size_arr, int size_work,
		      short compute_accuracy, factor_type allowed_factors)
{
#ifdef USE_SCSL
     typedef void scsl_fft_proc(int *,int*,int*,int*,FFTW_REAL*,
				FFTW_COMPLEX*,int*,int*,
				FFTW_COMPLEX*,int*,int*,
				FFTW_REAL*,FFTW_REAL*,int*);
     typedef void (*scsl_fft_proc_ptr)(int *,int*,int*,int*,FFTW_REAL*,
				       FFTW_COMPLEX*,int*,int*,
				       FFTW_COMPLEX*,int*,int*,
				       FFTW_REAL*,FFTW_REAL*,int*);
     scsl_fft_proc_ptr scsl_fft;
     extern scsl_fft_proc
	  FORTRANIZE(ccfft3d,CCFFT3D), FORTRANIZE(zzfft3d,ZZFFT3D);
     int isign, isign2 = +1, n1, n2, n3, ldx, ldy, ldx2, ldy2;
     FFTW_REAL scale = 1.0;
     FFTW_REAL *cray_fft_work, *cray_fft_table;
     int cray_fft_isys[1];
	  
     if (sizeof(FFTW_REAL) == sizeof(double))
	  scsl_fft = FORTRANIZE(zzfft3d,ZZFFT3D);
     else
	  scsl_fft = FORTRANIZE(ccfft3d,CCFFT3D);

     ldx = ldy = n1 = n_rev[0];
     ldx2 = ldy2 = n2 = n_rev[1];
     n3 = n_rev[2];
     
     cray_fft_isys[0] = 0;
     cray_fft_table = (FFTW_REAL*)work;
     cray_fft_work = cray_fft_table + (30 + 2 * n1) + (30 + 2 * n2) +
	                              (30 + 2 * n3);
     
     if (N != 0) {
	  isign = 0;  /* tells FFT we are initializing */	       
	  scsl_fft(&isign, &n1, &n2, &n3, &scale,
		   arr, &ldx, &ldx2, arr, &ldy, &ldy2,
		   cray_fft_table, cray_fft_work, cray_fft_isys);
     }
     isign = -1;
     
     DO_BENCHMARK_ND("SCSL",
		     3, n, N, arr, arr,
		     scsl_fft(&isign, &n1, &n2, &n3, &scale,
			      arr, &ldx, &ldx2, arr, &ldy, &ldy2,
			      cray_fft_table, cray_fft_work, 
			      cray_fft_isys),
		     1.0, -1, 1,
		     scsl_fft(&isign2, &n1, &n2, &n3, &scale,
			      arr, &ldx, &ldx2, arr, &ldy, &ldy2,
			      cray_fft_table, cray_fft_work, 
			      cray_fft_isys),
		     1.0/N,
		     compute_accuracy);
#endif
}

void do_imsl_3d_fft(int rank, int *n, int *n_rev, int N, 
		    short is_power_of_two,
		    FFTW_COMPLEX *arr, FFTW_COMPLEX *work,
		    int size_arr, int size_work,
		    short compute_accuracy, factor_type allowed_factors)
{
#ifdef USE_IMSL
     typedef void ffti_proc(int *,FFTW_COMPLEX *);
     typedef void fftfb_proc(int *,int *, int *,
			     FFTW_COMPLEX *,
			     int *, int *,
			     FFTW_COMPLEX *,
			     int *, int *,
			     FFTW_COMPLEX *,
			     FFTW_COMPLEX *,
			     FFTW_COMPLEX *,
			     FFTW_COMPLEX *);
     typedef void (*ffti_proc_ptr)(int *,FFTW_COMPLEX *);
     typedef void (*fftfb_proc_ptr)(int *,int *, int *,
				    FFTW_COMPLEX *,
				    int *, int *,
				    FFTW_COMPLEX *,
				    int *, int *,
				    FFTW_COMPLEX *,
				    FFTW_COMPLEX *,
				    FFTW_COMPLEX *,
				    FFTW_COMPLEX *);
     ffti_proc_ptr imsl_ffti;
     fftfb_proc_ptr imsl_fftf, imsl_fftb;
     extern ffti_proc FORTRANIZE(fftci,FFTCI), FORTRANIZE(dfftci,DFFTCI);
     extern fftfb_proc 
	  FORTRANIZE(f2t3f,F2T3F), FORTRANIZE(df2t3f,DF2T3F),
	  FORTRANIZE(f2t3b,F2T3B), FORTRANIZE(df2t3b,DF2T3B);
     extern int maxn(int n, int nums[]);
     FFTW_COMPLEX *w1,*w2,*w3;

     if (rank != 3) return;
    
     w1 = work + maxn(rank,n);
     w2 = w1 + 2*n_rev[0] + 8;
     w2 = w2 + 2*n_rev[1] + 8;

     if (sizeof(float) == sizeof(FFTW_REAL))  {
	  imsl_ffti = FORTRANIZE(fftci,FFTCI);
	  imsl_fftf = FORTRANIZE(f2t3f,F2T3F);
	  imsl_fftb = FORTRANIZE(f2t3b,F2T3B);
     }
     else {
	  imsl_ffti = FORTRANIZE(dfftci,DFFTCI);
	  imsl_fftf = FORTRANIZE(df2t3f,DF2T3F);
	  imsl_fftb = FORTRANIZE(df2t3b,DF2T3B);
     }
     
     if (N != 0) {
	  imsl_ffti(&n_rev[0], w1);
	  imsl_ffti(&n_rev[1], w2);
	  imsl_ffti(&n_rev[2], w3);
     }
     DO_BENCHMARK_ND("IMSL",
		     3, n, N, arr, arr,
		     imsl_fftf(&n_rev[0],&n_rev[1],&n_rev[2],
			       arr, &n_rev[0],&n_rev[1],
			       arr, &n_rev[0],&n_rev[1],
			       w1,w2,w3,work),
		     1.0, -1, 1,
		     imsl_fftb(&n_rev[0],&n_rev[1],&n_rev[2],
			       arr, &n_rev[0],&n_rev[1],
			       arr, &n_rev[0],&n_rev[1],
			       w1,w2,w3,work),
		     1.0/N,
		     compute_accuracy);     	  
#endif
}

void do_nag_3d_fft(int rank, int *n, int *n_rev, int N, 
		   short is_power_of_two,
		   FFTW_COMPLEX *arr, FFTW_COMPLEX *work,
		   int size_arr, int size_work,
		   short compute_accuracy, factor_type allowed_factors)
{
#ifdef USE_NAG
     typedef void fftfb_proc(int *ndim, int *n, int *n_tot,
			     FFTW_REAL *re, FFTW_REAL *im,
			     FFTW_REAL *work, int *lwork, int *fail);
     typedef void (*fftfb_proc_ptr)(int *ndim, int *n, int *n_tot,
				    FFTW_REAL *re, FFTW_REAL *im,
				    FFTW_REAL *work, int *lwork, int *fail);
     fftfb_proc_ptr nag_fftf;
     int fail = 0; /* MUST be initialized to 0 for NAG */
     FFTW_REAL *re,*im;
     int lwork = size_work / 2;
     extern fftfb_proc 
	  FORTRANIZE(c06fjf,C06FJF), FORTRANIZE(c06fje,C06FJE);

     if (rank != 3) return;
   
     if (sizeof(double) == sizeof(FFTW_REAL)) {
	  nag_fftf = FORTRANIZE(c06fjf,C06FJF);
     }
     else {
	  nag_fftf = FORTRANIZE(c06fje,C06FJE);
     }

     re = (FFTW_REAL *)arr;
     im = re + N;

     /* Note: NAG has no initialization or inverse routines. */
     
     if (N != 0) {
	  /* Call once to check for failures: */
	  nag_fftf(&rank,n_rev,&N,re,im,
		   (FFTW_REAL*)work,&lwork,&fail);
     }
     if (fail == 0) {
	  DO_BENCHMARK_ND("NAG",
			  3, n, N, arr, arr,
			  nag_fftf(&rank,n_rev,&N,re,im,
				   (FFTW_REAL*)work,&lwork,&fail),
			  sqrt(N), -1, 0,
			  0,
			  1.0/N,
			  -compute_accuracy);     	  
     }
     else {
	  if (fail == 1)
	       SKIP_BENCHMARK("rank < 1")
	  else if (fail == 2)
	       SKIP_BENCHMARK("N != product of dimensions")
	  else if (fail > 10 && fail % 10 == 1) 
               SKIP_BENCHMARK("NAG can't handle factors > 19")
	  else if (fail > 10 && fail % 10 == 2)
               SKIP_BENCHMARK("NAG can't handle more than 20 prime factors")
	  else if (fail > 10 && fail % 10 == 3)
	       SKIP_BENCHMARK("NAG can't handle N <= 1")
	  else if (fail > 10 && fail % 10 == 4)
	       SKIP_BENCHMARK("NAG requires work size >= 3 * max(n)")
	  else
	       SKIP_BENCHMARK("Unknown error in NAG")
     }
#endif
}

void do_dxml_3d_fft(int rank, int *n, int *n_rev, int N, 
		    short is_power_of_two,
		    FFTW_COMPLEX *arr, FFTW_COMPLEX *work,
		    int size_arr, int size_work,
		    short compute_accuracy, factor_type allowed_factors)
{
#if 0  /* Disabled because it doesn't seem to be working.  What is wrong? */
#ifdef USE_DXML
     if (rank != 3) return;
    
     if (sizeof(float) == sizeof(FFTW_REAL)) {
           int is_stride_1 = 1, stride = 1;
           DXML_C_FFT_STRUCTURE_3D fft_struct;

	   if (N != 0) cfft_init_3d_(&n_rev[0],&n_rev[1],&n_rev[2],
				     &fft_struct,&is_stride_1);
	   
	   DO_BENCHMARK_ND("DXML",
                     1, n, N, arr, arr,
                     cfft_apply_3d_("C","C","F",arr,arr,&n_rev[0],&n_rev[1],
				    &fft_struct,&stride,&stride,&stride),
                     1.0, -1, 1,
                     cfft_apply_3d_("C","C","B",arr,arr,&n_rev[0],&n_rev[1],
				    &fft_struct,&stride,&stride,&stride),
                     1.0/N,
                     compute_accuracy);

	   if (N != 0) cfft_exit_3d_(&fft_struct);
     }
     else if (sizeof(double) == sizeof(FFTW_REAL)) {
           int is_stride_1 = 1, stride = 1;
           DXML_Z_FFT_STRUCTURE_3D fft_struct;

	   if (N != 0) zfft_init_3d_(&n_rev[0],&n_rev[1],&n_rev[2],
				     &fft_struct,&is_stride_1);
	   
	   DO_BENCHMARK_ND("DXML",
                     1, n, N, arr, arr,
                     zfft_apply_3d_("C","C","F",arr,arr,&n_rev[0],&n_rev[1],
				    &fft_struct,&stride,&stride,&stride),
                     1.0, -1, 1,
                     zfft_apply_3d_("C","C","B",arr,arr,&n_rev[0],&n_rev[1],
				    &fft_struct,&stride,&stride,&stride),
                     1.0/N,
                     compute_accuracy);

	   if (N != 0) zfft_exit_3d_(&fft_struct);
     }
#endif
#endif
}
