/*
 * Copyright (c) 1997 Massachusetts Institute of Technology
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to use, copy, modify, and distribute the Software without
 * restriction, provided the Software, including any modified copies made
 * under this license, is not distributed for a fee, subject to
 * the following conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE MASSACHUSETTS INSTITUTE OF TECHNOLOGY BE LIABLE
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
 * CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 * Except as contained in this notice, the name of the Massachusetts
 * Institute of Technology shall not be used in advertising or otherwise
 * to promote the sale, use or other dealings in this Software without
 * prior written authorization from the Massachusetts Institute of
 * Technology.
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>

#include "bench_utils.h"

char which_fft[80] = "";
int which_fft_index = -1;

short bench_echo_dat = 1;
FILE *bench_log_file = 0,*bench_dat_file = 0;

int bench_num_transforms = 0, *bench_num_sizes = 0, bench_transform_index = 0;
int bench_t_index_real = 0;
double *bench_norm_avgs = 0, *bench_results = 0;

int check_prime_factors(int n, int maxprime)
/* returns 1 of the maximum prime factor of n is <= maxprime, and 0
   otherwise. */
{
     int primes[30] = { 2, 3, 5, 7, 11, 13, 17, 19, 23, 29, 31, 
			37, 41, 43, 47, 53, 59, 61, 67, 71, 73,
			79, 83, 89, 97, 101, 103, 107, 109, 113 };
     int i;

     if (maxprime > primes[30-1]) {
	  printf("\nBUG in check_prime_factors: maxprime is too big (%d)!\n",
		 maxprime);
	  exit(1);
     }
     if (n == 0)
	  return 1;
     for (i = 0; n > 1 && i < 30 && primes[i] <= maxprime; ++i) {
	  while (n % primes[i] == 0)
	       n /= primes[i];
     }
     return (n == 1);
}

void do_standard_fft(FFTW_COMPLEX *arr, int rank, int *size, int isign,
		     int reim_alt)
{
     int size_reverse[20]; /* GO expects array in Fortran order */
     int i, prod, N = 1;
     int go_fft(FFTW_REAL *a, FFTW_REAL *b, 
		int ntot, int n, int nspan, int isn);

     for (i = 0; i < rank; ++i) {
	  if (!check_prime_factors(size[i],23)) {
	       printf("\nSorry, the FFT we use to check for correctness\n"
		      "(Singleton) can't handle prime factors > 23.\n");
	       exit(1);
	  }
	  size_reverse[rank - 1 - i] = size[i];
	  N *= size[i];
     }
     
     /* FFT using GO */
     prod = 1;
     for (i = 0; i < rank; ++i) {
	  prod *= size_reverse[i];
	  if (reim_alt) /* alternating real/imag parts */
	       go_fft((FFTW_REAL *)arr, ((FFTW_REAL *)arr) + 1,
		      N, size_reverse[i], prod, 2 * isign);
	  else /* real/imag parts separate */
	       go_fft((FFTW_REAL *)arr, ((FFTW_REAL *)arr) + N,
		      N, size_reverse[i], prod, isign);
     }
}

void bench_init_array(FFTW_COMPLEX *arr, int N)
/* Initialize array arr of length N to arbitrary values. */
{
     int i = 0;
     while (i < N) {
	  c_re(arr[i]) = -1.3352;
	  c_im(arr[i++]) = 0.858274901;
     }
}

void bench_init_array_for_check(FFTW_COMPLEX *arr, int N)
/* Initialize array arr of length N to "random" values
   suitable for checking the accuracy of an FFT but
   reproducible. */
{
     int i;

     srand(3);

     for (i = 0; i < N; ++i) {
	  c_re(arr[i]) = rand() * 1.0e-9;
	  c_im(arr[i]) = rand() * 1.0e-9;
     }
}

double bench_check_array(FFTW_COMPLEX *arr, int N, double scale)
/* Returns mean error if array arr of length N.  arr[i] * scale
   should be the same as what bench_init_array_for_check produces;
   the mean error is the average of the absolute value of the
   differences. */
{
     double sum = 0.0, err;
     int i, n_errs_printed = 0;

     srand(3);

     for (i = 0; i < N; ++i) {
	  FFTW_REAL correct_im,correct_re;

	  correct_re = rand() * 1.0e-9;
	  correct_im = rand() * 1.0e-9;

	  err = fabs(scale*c_re(arr[i]) - correct_re)
	      + fabs(scale*c_im(arr[i]) - correct_im);
	  err *= 2.0;
	  err /= fabs(scale*c_re(arr[i])) +
	         fabs(correct_re) +
		 fabs(scale*c_im(arr[i])) + 
		 fabs(correct_im) + 1.0e-10;
	  sum += err;
	  if (bench_log_file && err > MAX_BENCH_ERROR*10 &&
	      n_errs_printed++ <= 10)
	       fprintf(bench_log_file,"\nFound big error at %d: \n"
		      "  (%g,%g)*%g = (%g,%g) != (%g,%g)\n", i,
		      c_re(arr[i]),c_im(arr[i]),scale,
		      scale*c_re(arr[i]),scale*c_im(arr[i]),
		      correct_re,
		      correct_im);
     }
     return (sum / N);
}

void bench_conjugate_array(FFTW_COMPLEX *arr, int n, short reim_alt)
/* Complex-conjugates the array arr[n].  reim_alt has the same meaning
   as the parameter passed to DO_BENCHMARK_ND: if reim_alt is 1, then
   arr is interpreted as a normal array of FFTW_COMPLEX numbers.  If
   reim_alt is zero, then arr is cast to an array of 2*n FFTW_REAL
   numbers--the first n numbers are the real parts (and are untouched)
   and the second n numbers are the imaginary parts (and are negated). */
{
     int i;

     if (reim_alt)
	  for (i = 0; i < n; ++i)
	       c_im(arr[i]) = -c_im(arr[i]);
     else {
	  FFTW_REAL *arr_im = ((FFTW_REAL *)arr) + n;
	  for (i = 0; i < n; ++i)
	       arr_im[i] = -arr_im[i];
     }
}

void bench_copy_array(FFTW_COMPLEX *from_arr, FFTW_COMPLEX *to_arr, int n)
{
     if (from_arr != to_arr) {
	  int i;
	  for (i = 0; i < n; ++i)
	       to_arr[i] = from_arr[i];
     }
}

/* The following procedures operate on global variables (ugh)
   defined in bench_utils.h. */

void init_normalized_averages(void)
{
     bench_t_index_real = bench_transform_index = bench_num_transforms = 0;
}

void compute_normalized_averages(int n, short compute_accuracy)
{
     if (compute_accuracy == 0 && OUTPUT_MFLOPS) {
          int i;
          if (n == 0) { /* initialize: */
	       if (bench_num_transforms == 0) {
		    printf("\n\nERROR!  No transforms in the benchmark!\n");
		    exit(1);
	       }

	       bench_num_sizes = fftw_malloc(sizeof(int) * bench_num_transforms);
	       bench_norm_avgs = fftw_malloc(sizeof(double)*bench_num_transforms);
	       bench_results = fftw_malloc(sizeof(double)*bench_num_transforms);
	       if (!bench_num_sizes || !bench_results || !bench_norm_avgs) {
	            printf("\n\nERROR!  Out of memory!\n");
		    exit(1);
	       }
	       for (i = 0; i < bench_num_transforms; ++i) {
	            bench_norm_avgs[i] = 0.0;
                    bench_num_sizes[i] = 0;
	       }
	       if (bench_log_file)
		 fprintf(bench_log_file,
			 "\nComputing normalized averages (%d transforms).\n",
			 bench_num_transforms);
	  }
	  else {
	       double max_result = -1;
	       for (i = 0; i < bench_num_transforms; ++i)
	            if (bench_results[i] > max_result)
		         max_result = bench_results[i];
	       for (i = 0; i < bench_num_transforms; ++i)
		    if (bench_results[i] >= 0.0)
			 bench_norm_avgs[i] += bench_results[i] / max_result;
	       if (bench_log_file) {
		    fprintf(bench_log_file,
			    "Top mflops for N=%d = %g\n",n,max_result);
	            fprintf(bench_log_file,
			    "Normalized results and averages for N=%d:\n",n);
		    for (i = 0; i < bench_num_transforms; ++i)
		         fprintf(bench_log_file,
				 "  fft %d: mflops = %g (norm. = %g)"
				 ", norm. avg. (of %d) = %g\n",
				 i, bench_results[i], 
				 bench_results[i] / max_result,
				 bench_num_sizes[i],
				 bench_num_sizes[i] ? 
				 bench_norm_avgs[i] / bench_num_sizes[i]
				 : -1.0);
		    fflush(bench_log_file);
	       }
	  }
          bench_t_index_real = bench_transform_index = 0;
          for (i = 0; i < bench_num_transforms; ++i)
               bench_results[i] = 0.0;
     }
     else
          bench_t_index_real=bench_transform_index = bench_num_transforms = 0;
}

void output_normalized_averages(void)
{
     if (bench_num_transforms > 0 && bench_norm_avgs && bench_num_sizes) {
          int i;
	  if (bench_dat_file) {
	       fprintf(bench_dat_file,"Norm. Avg.");
	       for (i = 0; i < bench_num_transforms; ++i)
		    if (bench_num_sizes[i] > 0)
			 fprintf(bench_dat_file,", %g", 
				 bench_norm_avgs[i] / bench_num_sizes[i]);
		    else
			 fprintf(bench_dat_file,", ");
	       fprintf(bench_dat_file,"\n");
	  }
	  if (bench_echo_dat) {
	       printf("Norm. Avg.");
	       for (i = 0; i < bench_num_transforms; ++i)
		    if (bench_num_sizes[i] > 0)
			 printf(", %g", 
				 bench_norm_avgs[i] / bench_num_sizes[i]);
		    else
			 printf(", ");
	       printf("\n");
	  }
	  bench_num_transforms = 0;
	  if (bench_num_sizes) fftw_free(bench_num_sizes);
	  if (bench_norm_avgs) fftw_free(bench_norm_avgs);
	  if (bench_results) fftw_free(bench_results);
	  bench_num_sizes = 0;
	  bench_norm_avgs = 0;
	  bench_results = 0;
     }
}

