/*
 * Copyright (c) 1997 Massachusetts Institute of Technology
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to use, copy, modify, and distribute the Software without
 * restriction, provided the Software, including any modified copies made
 * under this license, is not distributed for a fee, subject to
 * the following conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE MASSACHUSETTS INSTITUTE OF TECHNOLOGY BE LIABLE
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
 * CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 * Except as contained in this notice, the name of the Massachusetts
 * Institute of Technology shall not be used in advertising or otherwise
 * to promote the sale, use or other dealings in this Software without
 * prior written authorization from the Massachusetts Institute of
 * Technology.
 *
 */

#ifndef BENCH_UTILS_H
#define BENCH_UTILS_H

#include <fftw.h>

/* Check if we are using FFTW 1.3; if so, we need to include fftw-int.h.
   The easiest way to do this is to check for count_plan_ops, which was new
   in FFTW 1.3: */
#ifdef FFTW_HAS_COUNT_PLAN_OPS
#include <fftw-int.h>
#endif

#include <stdio.h>
#include <string.h>

extern char which_fft[80]; /* which FFT to benchmark, or "" for all */
extern int which_fft_index; /* index of FFT corresponding to which_fft,
			       or -1 if which_fft is "" */

extern short bench_echo_dat;
extern FILE *bench_log_file,*bench_dat_file;

#define OUTPUT_MFLOPS 1  /* if 0, outputs time/(n log2 n) */

typedef enum {
     POWERS_OF_TWO_ONLY,
     NON_POWERS_OF_TWO_ONLY,
     ALL_FACTORS
} factor_type;

short bench_1d(short compute_accuracy, 
	       factor_type allowed_factors, int which_N, double max_MB);
short bench_3d(short compute_accuracy, 
	       factor_type allowed_factors, int which_N, double max_MB);

void bench_init_array(FFTW_COMPLEX *arr, int N);
void bench_init_array_for_check(FFTW_COMPLEX *arr, int N);
double bench_check_array(FFTW_COMPLEX *arr, int N, double scale);

void bench_conjugate_array(FFTW_COMPLEX *arr, int n, short reim_alt);
void bench_copy_array(FFTW_COMPLEX *from_arr, FFTW_COMPLEX *to_arr, int n);

/*
 * convert C name -> FORTRAN name.  On some systems, 
 * append an underscore. On other systems, use all caps.
 *
 * x is the lower case name, X is the all caps name.
 */

#if defined(CRAY) || defined(_UNICOS) || defined(_CRAYMPP)
#define FORTRANIZE(x,X) X /* all upper-case on the Cray */

#elif defined(SOLARIS)
#define FORTRANIZE(x,X) x##_  /* append an underscore for Solaris */

#elif defined(IBM6000) || defined(_AIX)
#define FORTRANIZE(x,X) x  /* all lower-case on RS/6000 */

#else
#define FORTRANIZE(x,X) x##_  /* use all lower-case with 
				 appended underscore by default */

#endif

#ifdef USE_FORTRAN
#define WHEN_FORTRAN(foobar) foobar
#else
#define WHEN_FORTRAN(foobar) 0
#endif

#define IS_POWER_OF_TWO(n) ((((n)-1) & ~(n)) + 1 == (n))

#define MAX_BENCH_ERROR 1.0e-3
#define MIN_BENCH_TIME 1.0

#ifndef FFTW_HAS_TIME_DIFF /* old version of fftw, no fftw_time_diff */
#define fftw_time_diff(t1,t2) ((t1) - (t2))
#endif

extern int bench_num_transforms, bench_transform_index, bench_t_index_real;
extern int *bench_num_sizes;
extern double *bench_norm_avgs, *bench_results;

void init_normalized_averages(void);
void compute_normalized_averages(int n, short compute_accuracy);
void output_normalized_averages(void);

void do_standard_fft(FFTW_COMPLEX *arr, int rank, int *size, int sign,
		     int reim_alt);

/* Macro to do the ND benchmark:

   Parameters:
   name = name of FFT
   rank = rank (dimensionality) of transform
   size[rank] = size of transform (array of integers); if product of
                sizes is zero, then print name of transform instead
   real_N = N to normalize time by (usually should be product of
            sizes, but may be different for CWP)
   in = input array
   out = output array (can equal in)
   fft = command to transform "in"
   scale_1 = scale factor to multiply by after fft to get unnormalized FFT
   sign = sign (+1/-1) in exponent of fft
   reim_alt = 1 if real & imag. parts are stored in alternating order,
              and 0 if they are stored separately (in consecutive, contiguous
	      chunks)...this is used to check the correctness of the FFT
	      using GO; set to -1 if you don't want to check at all
   ifft = command to perform inverse of fft, putting results in "in"
   scale = scale factor to multiply by after ifft to get true inverse
   acc = whether to compute timing (0) or acc (1 if have ifft, -1 otherwise)
*/
#define DO_BENCHMARK_ND(name,rank,size,real_N,in,out,fft,scale_1,sign,reim_alt,ifft,scale,acc) {\
     fftw_time start_t,end_t; \
     double mean_error, t, mflops, bench_res; \
     int iter, iters = 1, N_tot = 1; \
     if (which_fft[0] == 0 || !strcmp(which_fft,name)) { \
     if (which_fft[0] != 0) which_fft_index = bench_t_index_real; \
     for (iter = 0; iter < rank; ++iter) N_tot *= (size)[iter]; \
     if (N_tot == 0) { \
	  if (bench_dat_file) fprintf(bench_dat_file,", %s",name); \
	  if (bench_echo_dat) printf(", %s",name); \
          if (bench_log_file) \
	       fprintf(bench_log_file,"%d. %s\n",bench_num_transforms,name); \
	  ++bench_num_transforms; \
          ++bench_transform_index; \
     } \
     else { \
          if (bench_log_file) { \
	       fprintf(bench_log_file,"%d. %s",bench_transform_index,name); \
               if (N_tot == (real_N)) \
	            fprintf(bench_log_file,": "); \
	       else \
	            fprintf(bench_log_file," (N=%d): ",N_tot); \
	  } \
	  if (acc) { \
	       bench_init_array_for_check(in,N_tot); \
	       fft; \
	       if (acc < 0) { \
		    bench_conjugate_array(out,N_tot,reim_alt); \
		    bench_copy_array(out,in,N_tot); \
		    fft; \
		    bench_conjugate_array(out,N_tot,reim_alt); \
	            mean_error = bench_check_array(out,N_tot,           \
						   (scale_1)*(scale_1)  \
						   * 1.0/N_tot);        \
	       } \
               else { \
		    ifft; \
	            mean_error = bench_check_array(in,N_tot,scale); \
	       } \
               if (bench_log_file) \
		    fprintf(bench_log_file, \
			    " mean fractional error = %e\n", \
			    mean_error); \
	       if (bench_dat_file) { \
                    fprintf(bench_dat_file,", %e",mean_error); \
                    fflush(bench_dat_file); \
	       } \
	       if (bench_echo_dat) { \
                    printf(", %e",mean_error); \
                    fflush(stdout); \
	       } \
               ++bench_transform_index; \
	  } \
	  else { \
	       bench_init_array(in,N_tot); \
	       fft; \
	       do { \
		    start_t = fftw_get_time(); \
		    for (iter = 0; iter < iters; ++iter) { \
			 bench_init_array(in,N_tot); \
			 fft; \
		    } \
		    end_t = fftw_get_time(); \
		    t = fftw_time_to_sec(fftw_time_diff(end_t,start_t)); \
		    iters *= 2; \
	       } while (t < MIN_BENCH_TIME); \
	       iters /= 2; \
	       if (bench_log_file) \
		    fprintf(bench_log_file,                           \
			    "elapsed time t=%g s, %d iters",t,iters); \
	       start_t = fftw_get_time(); \
	       for (iter = 0; iter < iters; ++iter) \
		    bench_init_array(in,N_tot); \
	       end_t = fftw_get_time(); \
	       t -= fftw_time_to_sec(fftw_time_diff(end_t,start_t)); \
               if (bench_log_file) \
		    fprintf(bench_log_file,", t-(init.)=%g s\n",t); \
	       t /= 1e-6 * (real_N) * log(1.0*(real_N))/log(2.0) * iters; \
	       mflops = 5.0 / t; \
               bench_res = OUTPUT_MFLOPS ? mflops : t; \
               if (bench_log_file) \
		    fprintf(bench_log_file, \
			    "    t(norm)=%g, mflops=%g",t,mflops); \
	       if (bench_dat_file) { \
                    fprintf(bench_dat_file,", %g",bench_res); \
                    fflush(bench_dat_file); \
	       } \
               if (bench_echo_dat) { \
                    printf(", %g",bench_res); \
                    fflush(stdout); \
               } \
               if (bench_results) \
                    bench_results[bench_transform_index] = bench_res; \
               if (bench_num_sizes) \
                    bench_num_sizes[bench_transform_index] += 1; \
               ++bench_transform_index; \
	       if (reim_alt != -1) { \
	       bench_init_array_for_check(in,N_tot); \
	       fft; \
               do_standard_fft(out,rank,size,-(sign),reim_alt); \
	       mean_error = bench_check_array(out,N_tot,(scale_1)*1.0/N_tot); \
               if (bench_log_file) \
		    fprintf(bench_log_file," (err=%0.1e)",mean_error); \
	       if (mean_error > MAX_BENCH_ERROR) { \
		    fprintf(bench_log_file, \
			    "\n\n%s gave wrong answer! (mean err = %g)\n", \
			    name,mean_error); \
		    printf("\n\n%s gave wrong answer! (mean err = %g)\n", \
			   name,mean_error); \
		    exit(1); \
	       } \
	       }\
               if (bench_log_file) { \
		    fprintf(bench_log_file,"\n"); \
                    fflush(bench_log_file); \
	       } \
	  } \
     } \
     } \
     bench_t_index_real++; \
}

#define SKIP_BENCHMARK(why) { \
     if (which_fft_index == -1 || bench_t_index_real == which_fft_index) { \
     if (bench_dat_file) \
	  fprintf(bench_dat_file,", "); \
     if (bench_echo_dat) \
	  printf(", "); \
     if (bench_log_file) \
          fprintf(bench_log_file,"%d. Skipping fft (%s).\n", \
                  bench_transform_index, why); \
     if (bench_num_transforms) bench_results[bench_transform_index] = -1; \
     ++bench_transform_index; \
     } \
     bench_t_index_real++; \
}

#endif
